#!/usr/bin/env python
# crate_anon/crateweb/userprofile/views.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from django.http import HttpResponse
from django.http.request import HttpRequest
from django.shortcuts import redirect, render
from crate_anon.crateweb.userprofile.forms import UserProfileForm


# =============================================================================
# User profile settings
# =============================================================================
# http://www.slideshare.net/pydanny/advanced-django-forms-usage
# ... e.g. slide 72

def edit_profile(request: HttpRequest) -> HttpResponse:
    profile = request.user.profile
    form = UserProfileForm(request.POST if request.method == 'POST' else None,
                           instance=profile)
    if form.is_valid():
        profile = form.save()
        profile.save()
        return redirect('home')
    return render(request, 'edit_profile.html',
                  {'form': form, 'profile': profile})



#!/usr/bin/env python
# crate_anon/anonymise/patient.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Patient class for CRATE anonymiser.
"""

import logging
from typing import AbstractSet, Any, Generator, List

from sqlalchemy.sql import column, select, table

from crate_anon.anonymise.config_singleton import config
from crate_anon.anonymise.constants import SCRUBSRC
from crate_anon.anonymise.models import PatientInfo
from crate_anon.anonymise.scrub import PersonalizedScrubber

log = logging.getLogger(__name__)


# =============================================================================
# Generate identifiable values for a patient
# =============================================================================

def gen_all_values_for_patient(dbname: str,
                               tablename: str,
                               fields: List[str],
                               pid: int) -> Generator[List[Any], None, None]:
    """
    Generate all sensitive (scrub_src) values for a given patient, from a given
    source table. Used to build the scrubber.

        sources: dictionary
            key: db name
            value: rnc_db database object
        dbname: source database name
        table: source table
        fields: source fields containing scrub_src information
        pid: patient ID

    Yields rows, where each row is a list of values that matches "fields".
    """
    cfg = config.sources[dbname].srccfg
    if not cfg.ddgen_per_table_pid_field:
        return
        # http://stackoverflow.com/questions/13243766
    log.debug(
        "gen_all_values_for_patient: PID {p}, table {d}.{t}, "
        "fields: {f}".format(
            d=dbname, t=tablename, f=",".join(fields), p=pid))
    session = config.sources[dbname].session
    query = (
        select([column(f) for f in fields]).
        where(column(cfg.ddgen_per_table_pid_field) == pid).
        select_from(table(tablename))
    )
    result = session.execute(query)
    for row in result:
        log.debug("... yielding row: {}".format(row))
        yield row


# =============================================================================
# Patient class, which hosts the patient-specific scrubber
# =============================================================================

class Patient(object):
    """Class representing a patient-specific information, such as PIDs, RIDs,
    and scrubbers."""

    def __init__(self, pid: int, debug: bool = False) -> None:
        """
        Build the scrubber based on data dictionary information.

            sources: dictionary
                key: db name
                value: rnc_db database object
            pid: integer patient identifier
        """
        self.pid = pid
        self.session = config.admindb.session

        # Fetch or create PatientInfo object
        self.info = self.session.query(PatientInfo).get(pid)
        if self.info is None:
            self.info = PatientInfo(pid=pid)
            self.info.ensure_rid()
            self.info.ensure_trid(self.session)
            self.session.add(self.info)
            self.session.commit()
            # prompt commit after insert operations, to ensure no locks

        # Scrubber
        self.scrubber = PersonalizedScrubber(
            anonymise_codes_at_word_boundaries_only=(
                config.anonymise_codes_at_word_boundaries_only),
            anonymise_dates_at_word_boundaries_only=(
                config.anonymise_dates_at_word_boundaries_only),
            anonymise_numbers_at_word_boundaries_only=(
                config.anonymise_numbers_at_word_boundaries_only),
            anonymise_numbers_at_numeric_boundaries_only=(
                config.anonymise_numbers_at_numeric_boundaries_only),
            anonymise_strings_at_word_boundaries_only=(
                config.anonymise_strings_at_word_boundaries_only),
            debug=debug,
            hasher=config.change_detection_hasher,
            min_string_length_for_errors=config.min_string_length_for_errors,
            min_string_length_to_scrub_with=(
                config.min_string_length_to_scrub_with),
            nonspecific_scrubber=config.nonspecific_scrubber,
            replacement_text_patient=config.replace_patient_info_with,
            replacement_text_third_party=config.replace_third_party_info_with,
            scrub_string_suffixes=config.scrub_string_suffixes,
            string_max_regex_errors=config.string_max_regex_errors,
            whitelist=config.whitelist,
        )
        # Database
        # Construction. We go through all "scrub-from" fields in the data
        # dictionary. We collect all values of those fields from the source
        # database.
        log.debug("Building scrubber")
        self._db_table_pair_list = config.dd.get_scrub_from_db_table_pairs()
        self._mandatory_scrubbers_unfulfilled = \
            config.dd.get_mandatory_scrubber_sigs().copy()
        self._build_scrubber(pid,
                             depth=0,
                             max_depth=config.thirdparty_xref_max_depth)
        self._unchanged = self.get_scrubber_hash() == self.info.scrubber_hash
        self.info.set_scrubber_info(self.scrubber)
        self.session.commit()
        # Commit immediately, because other processes may need this table
        # promptly. Otherwise, might get:
        #   Deadlock found when trying to get lock; try restarting transaction

    def _build_scrubber(self, pid: int, depth: int, max_depth: int) -> None:
        if depth > 0:
            log.debug("Building scrubber recursively: depth = {}".format(
                depth))
        for (src_db, src_table) in self._db_table_pair_list:
            # Build a list of fields for this table.
            ddrows = config.dd.get_scrub_from_rows(src_db, src_table)
            fields = [ddr.src_field for ddr in ddrows]
            # Precalculate things; we might being going through a lot of values
            scrub_method = [
                PersonalizedScrubber.get_scrub_method(ddr.src_datatype,
                                                      ddr.scrub_method)
                for ddr in ddrows
            ]
            is_patient = [depth == 0 and ddr.scrub_src is SCRUBSRC.PATIENT
                          for ddr in ddrows]
            is_mpid = [depth == 0 and ddr.master_pid for ddr in ddrows]
            recurse = [depth < max_depth and
                       ddr.scrub_src is SCRUBSRC.THIRDPARTY_XREF_PID
                       for ddr in ddrows]
            required_scrubber = [ddr.required_scrubber for ddr in ddrows]
            sigs = [ddr.get_signature() for ddr in ddrows]
            # Collect the actual patient-specific values for this table.
            for values in gen_all_values_for_patient(src_db, src_table,
                                                     fields, pid):
                for i, val in enumerate(values):
                    self.scrubber.add_value(val, scrub_method[i],
                                            patient=is_patient[i])

                    if is_mpid[i] and self.get_mpid() is None:
                        # We've come across the master ID.
                        self.set_mpid(val)

                    if recurse[i]:
                        # We've come across a patient ID of another patient,
                        # whose information should be trawled and treated
                        # as third-party information
                        try:
                            related_pid = int(val)
                        except (ValueError, TypeError):
                            # TypeError: NULL value (None)
                            # ValueError: duff value, i.e. non-integer
                            continue
                        self._build_scrubber(related_pid, depth + 1, max_depth)

                    if val is not None and required_scrubber[i]:
                        self._mandatory_scrubbers_unfulfilled.discard(sigs[i])

    @property
    def mandatory_scrubbers_unfulfilled(self) -> AbstractSet[str]:
        return self._mandatory_scrubbers_unfulfilled

    def get_pid(self) -> int:
        """Return the patient ID (PID)."""
        return self.info.pid

    def get_mpid(self) -> int:
        """Return the master patient ID (MPID)."""
        return self.info.mpid

    def set_mpid(self, mpid: int) -> None:
        self.info.set_mpid(mpid)

    def get_rid(self) -> str:
        """Returns the RID (encrypted PID)."""
        return self.info.rid

    def get_mrid(self) -> str:
        """Returns the master RID (encrypted MPID)."""
        return self.info.mrid

    def get_trid(self) -> int:
        """Returns the transient integer RID (TRID)."""
        return self.info.trid

    def get_scrubber_hash(self) -> str:
        return self.scrubber.get_hash()

    def scrub(self, text: str) -> str:
        return self.scrubber.scrub(text)

    def unchanged(self) -> bool:
        """
        Has the scrubber changed, compared to the previous hashed version in
        the admin database?
        """
        return self._unchanged



#!/usr/bin/env python
# crate_anon/crateweb/core/admin.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

# NOTE that
# - Objects for which a user is not authorized, via get_queryset(), will
#   causes an Http404 error.
# - You can't filter on Python properties via the Django QuerySet ORM.

import logging
from typing import Any, Dict, Iterable, Tuple

from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.contrib.auth.models import User
from django.contrib.auth.admin import UserAdmin
from django.core.urlresolvers import reverse
from django.db import transaction
from django.db.models import Q, QuerySet
from django.http.request import HttpRequest
from django.template.defaultfilters import yesno
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy
from crate_anon.crateweb.extra.admin import (
    AddOnlyModelAdmin,
    admin_view_fk_link,
    admin_view_reverse_fk_links,
    AllStaffReadOnlyModelAdmin,
    EditOnceOnlyModelAdmin,
    EditOnlyModelAdmin,
    disable_bool_icon,
    ReadOnlyModelAdmin,
)
from crate_anon.common.stringfunc import replace_in_list
from crate_anon.crateweb.userprofile.models import UserProfile
from crate_anon.crateweb.consent.models import (
    CharityPaymentRecord,
    ClinicianResponse,
    ConsentMode,
    ContactRequest,
    Decision,
    DummyPatientSourceInfo,
    Email,
    Leaflet,
    Letter,
    PatientLookup,
    PatientResponse,
    Study,
    TeamRep,
)
from crate_anon.crateweb.consent.tasks import (
    process_consent_change,
    process_patient_response,
    resend_email,
)
from crate_anon.crateweb.research.models import (
    QueryAudit,
)

log = logging.getLogger(__name__)


# =============================================================================
# Research
# =============================================================================

class QueryMgrAdmin(ReadOnlyModelAdmin):
    model = QueryAudit
    # Make all fields read-only (see also ReadOnlyModelAdmin):
    readonly_fields = ('id', 'when', 'get_user', 'get_sql', 'get_count_only',
                       'n_records', 'get_failed', 'fail_msg')
    fields = readonly_fields  # or other things could appear
    # Group entries by date conveniently:
    date_hierarchy = 'when'
    # Prefetch related objects (hugely reduces number of SQL queries):
    list_select_related = ('query', 'query__user')
    # What to show in the list:
    list_display = ('id', 'when', 'get_user', 'get_sql', 'get_count_only',
                    'n_records', 'get_failed', 'fail_msg')
    # Filter on Booleans on the right-hand side:
    list_filter = ('count_only', 'failed')
    # Search text content of these:
    search_fields = ('query__sql', 'query__user__username')

    def get_sql(self, obj: QueryAudit) -> str:
        return obj.query.sql
    get_sql.short_description = "SQL"
    get_sql.admin_order_field = 'query__sql'

    def get_user(self, obj: QueryAudit) -> str:
        return obj.query.user
    get_user.short_description = "User"
    get_user.admin_order_field = 'query__user'

    def get_count_only(self, obj: QueryAudit) -> str:
        return yesno(obj.count_only)
    get_count_only.short_description = "Count only?"
    get_count_only.admin_order_field = 'count_only'

    def get_failed(self, obj: QueryAudit) -> str:
        return yesno(obj.failed)
    get_failed.short_description = "Failed?"
    get_failed.admin_order_field = 'failed'


# =============================================================================
# Consent
# =============================================================================

# -----------------------------------------------------------------------------
# Study
# -----------------------------------------------------------------------------

class StudyInline(admin.TabularInline):
    model = Study


class StudyMgrAdmin(admin.ModelAdmin):
    fields = (
        'institutional_id',
        'title', 'lead_researcher', 'registered_at', 'summary',
        'search_methods_planned', 'patient_contact', 'include_under_16s',
        'include_lack_capacity', 'clinical_trial', 'include_discharged',
        'request_direct_approach', 'approved_by_rec', 'rec_reference',
        'approved_locally', 'local_approval_at',
        'study_details_pdf', 'subject_form_template_pdf',
        'researchers',
    )
    list_display = ('id', 'institutional_id', 'title', 'lead_researcher')
    list_display_links = ('id', 'institutional_id', 'title')
    filter_horizontal = ('researchers', )


class StudyResAdmin(AllStaffReadOnlyModelAdmin):
    fields = (
        'institutional_id',
        'title', 'lead_researcher', 'registered_at', 'summary',
        'search_methods_planned', 'patient_contact', 'include_under_16s',
        'include_lack_capacity', 'clinical_trial', 'include_discharged',
        'request_direct_approach', 'approved_by_rec', 'rec_reference',
        'approved_locally', 'local_approval_at',
        'study_details_pdf', 'subject_form_template_pdf',
        'researchers',
    )
    readonly_fields = fields
    list_display = ('id', 'institutional_id', 'title', 'lead_researcher')
    list_display_links = ('id', 'institutional_id', 'title')

    # Restrict to studies that this researcher is affiliated to
    def get_queryset(self, request: HttpRequest) -> QuerySet:
        qs = super().get_queryset(request)
        return Study.filter_studies_for_researcher(qs, request.user)


# -----------------------------------------------------------------------------
# Leaflet
# -----------------------------------------------------------------------------

class LeafletMgrAdmin(EditOnlyModelAdmin):
    fields = ('name', 'pdf')
    readonly_fields = ('name', )


class LeafletResAdmin(AllStaffReadOnlyModelAdmin):
    fields = ('name', 'get_pdf')
    readonly_fields = fields

    def get_pdf(self, obj: Leaflet) -> str:
        if not obj.pdf:
            return "(Missing)"
        return '<a href="{}">PDF</a>'.format(
            reverse("leaflet", args=[obj.name]))
    get_pdf.short_description = "Leaflet PDF"
    get_pdf.allow_tags = True


# -----------------------------------------------------------------------------
# E-mail
# -----------------------------------------------------------------------------

class EmailSentListFilter(SimpleListFilter):
    title = 'email sent'
    parameter_name = 'email_sent'

    def lookups(self,
                request: HttpRequest,
                model_admin: admin.ModelAdmin) -> Iterable[Tuple[str, str]]:
        return (
            ('y', "E-mail sent at least once"),
            ('n', "E-mail not sent"),
        )

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        if self.value() == 'y':
            return queryset.filter(emailtransmission__sent=True).distinct()
        if self.value() == 'n':
            return queryset.exclude(emailtransmission__sent=True)


class EmailDevAdmin(ReadOnlyModelAdmin):
    model = Email
    readonly_fields = (
        'id', 'created_at', 'sender', 'recipient', 'subject',
        'msg_text', 'get_view_msg_html', 'get_view_attachments',
        'to_clinician', 'to_researcher', 'to_patient',
        'get_study', 'get_contact_request', 'get_letter',
        'get_sent', 'get_transmissions',
    )
    fields = readonly_fields  # or other things appear
    date_hierarchy = 'created_at'
    list_display = ('id', 'created_at', 'recipient', 'subject', 'get_sent')
    list_filter = (EmailSentListFilter, )
    search_fields = ('recipient', 'subject')
    actions = ['resend']
    # ... alternative method (per instance):
    # http://stackoverflow.com/questions/2805701/is-there-a-way-to-get-custom-django-admin-actions-to-appear-on-the-change-view  # noqa

    # - We can't use list_select_related for things that have a foreign key to
    #   us (rather than things we have an FK to).
    # - prefetch_related (in the queryset) just uses Python, not SQL.
    # - http://blog.roseman.org.uk/2010/01/11/django-patterns-part-2-efficient-reverse-lookups/  # noqa
    # Anyway, premature optimization is the root of all evil, and all that.

    def get_view_msg_html(self, obj: Email) -> str:
        url = reverse('view_email_html', args=[obj.id])
        return '<a href="{}">View HTML message</a> ({} bytes)'.format(
            url, len(obj.msg_html))
    get_view_msg_html.short_description = "Message HTML"
    get_view_msg_html.allow_tags = True

    def get_view_attachments(self, obj: Email) -> str:
        attachments = obj.emailattachment_set.all()
        if not attachments:
            return "(No attachments)"
        html = ""
        for i, attachment in enumerate(attachments):
            if attachment.exists():
                html += (
                    'Attachment {}: <a href="{}"><b>{}</b></a> '
                    '({} bytes), sent as <b>{}</b></a><br>'
                ).format(
                    i + 1,
                    reverse('view_email_attachment', args=[attachment.id]),
                    attachment.file,
                    attachment.size(),
                    attachment.sent_filename,
                )
            else:
                html += (
                    'Attachment {}: <b>{}</b> (missing), '
                    'sent as <b>{}</b><br>'
                ).format(
                    i + 1,
                    attachment.file,
                    attachment.sent_filename,
                )
        return html
    get_view_attachments.short_description = "Attachments"
    get_view_attachments.allow_tags = True

    def resend(self, request: HttpRequest, queryset: QuerySet) -> None:
        email_ids = []
        for email in queryset:
            email_ids.append(email.id)
            # transaction.on_commit not required (no changes made to emails)
            resend_email.delay(email.id, request.user.id)  # Asynchronous
        if email_ids:
            self.message_user(
                request,
                "{} e-mails were queued for resending: IDs {}.".format(
                    len(email_ids), str(email_ids)))
    resend.short_description = "Resend selected e-mails"

    def get_transmissions(self, obj: Email) -> str:
        return "<br>".join(str(x) for x in obj.emailtransmission_set.all())
    get_transmissions.short_description = "Transmissions"
    get_transmissions.allow_tags = True

    def get_sent(self, obj: Email) -> bool:
        return obj.has_been_sent()
    get_sent.short_description = "Sent"
    get_sent.boolean = True

    def get_letter(self, obj: Email) -> str:
        return admin_view_fk_link(self, obj, "letter")
    get_letter.short_description = "Letter"
    get_letter.allow_tags = True

    def get_study(self, obj: Email) -> str:
        return admin_view_fk_link(self, obj, "study")
    get_study.short_description = "Study"
    get_study.allow_tags = True

    def get_contact_request(self, obj: Email) -> str:
        return admin_view_fk_link(self, obj, "contact_request")
    get_contact_request.short_description = "Contact request"
    get_contact_request.allow_tags = True


class EmailMgrAdmin(EmailDevAdmin):
    """
    Restrict to e-mails/information visible to the RDBM.
    Also, since we're not inhering from AllStaffReadOnlyModelAdmin, give
    admin read permissions to all staff.
    """
    readonly_fields = (
        'id', 'created_at', 'sender', 'recipient', 'subject',

        # subject should not be confidential
        # no text, HTML, or attachments
        'get_restricted_msg_text', 'get_restricted_msg_html',
        'get_restricted_attachments',

        'to_clinician', 'to_researcher', 'to_patient',
        'get_study', 'get_contact_request', 'get_letter',
        'get_sent', 'get_transmissions',
    )
    fields = readonly_fields  # or other things appear
    actions = ['resend']

    def get_queryset(self, request: HttpRequest) -> QuerySet:
        qs = super().get_queryset(request).filter(Q(to_researcher=True) |
                                                  Q(to_patient=True))
        return qs

    @staticmethod
    def rdbm_may_view(obj: Email) -> bool:
        return obj.to_patient or obj.to_researcher

    def get_restricted_msg_text(self, obj: Email) -> str:
        if not self.rdbm_may_view(obj):
            return "(Not authorized)"
        return obj.msg_text
    get_restricted_msg_text.short_description = "Message text"

    def get_restricted_msg_html(self, obj: Email) -> str:
        if not self.rdbm_may_view(obj):
            return "(Not authorized)"
        return self.get_view_msg_html(obj)
    get_restricted_msg_html.short_description = "Message HTML"
    get_restricted_msg_html.allow_tags = True

    def get_restricted_attachments(self, obj: Email) -> str:
        if not self.rdbm_may_view(obj):
            return "(Not authorized)"
        return self.get_view_attachments(obj)
    get_restricted_attachments.short_description = "Attachments"
    get_restricted_attachments.allow_tags = True


class EmailResAdmin(EmailDevAdmin):
    """
    Restrict to e-mails visible to a researcher.
    Also, since we're not inhering from AllStaffReadOnlyModelAdmin, give
    admin read permissions to all staff.
    """
    readonly_fields = (
        'id', 'created_at', 'sender', 'recipient', 'subject',
        'msg_text', 'get_view_msg_html', 'get_view_attachments',
        'to_clinician', 'to_researcher', 'to_patient',
        'get_study', 'get_contact_request', 'get_letter',
        'get_sent', 'get_transmissions',
    )
    fields = readonly_fields  # or other things appear
    actions = None  # not [], which allows site-wide things

    def get_queryset(self, request: HttpRequest) -> QuerySet:
        qs = super().get_queryset(request).filter(to_researcher=True)
        studies = Study.filter_studies_for_researcher(Study.objects.all(),
                                                      request.user)
        return qs.filter(study__in=studies)

    def has_module_permission(self, request: HttpRequest) -> bool:
        return request.user.is_staff

    def has_change_permission(self,
                              request: HttpRequest,
                              obj: Email = None) -> bool:
        return request.user.is_staff


# -----------------------------------------------------------------------------
# Dummy patient source info
# -----------------------------------------------------------------------------

class DummyPatientSourceInfoDevAdmin(admin.ModelAdmin):
    fields = (
        # Patient
        'nhs_number',
        'pt_dob', 'pt_dod', 'pt_dead', 'pt_discharged', 'pt_discharge_date',
        'pt_sex',
        'pt_title', 'pt_first_name', 'pt_last_name',
        'pt_address_1', 'pt_address_2', 'pt_address_3', 'pt_address_4',
        'pt_address_5', 'pt_address_6', 'pt_address_7',
        'pt_telephone', 'pt_email',

        # GP
        'gp_title', 'gp_first_name', 'gp_last_name',
        'gp_address_1', 'gp_address_2', 'gp_address_3', 'gp_address_4',
        'gp_address_5', 'gp_address_6', 'gp_address_7',
        'gp_telephone', 'gp_email',

        # Clinician
        'clinician_title', 'clinician_first_name', 'clinician_last_name',
        'clinician_address_1', 'clinician_address_2', 'clinician_address_3',
        'clinician_address_4', 'clinician_address_5', 'clinician_address_6',
        'clinician_address_7',
        'clinician_telephone', 'clinician_email',
        'clinician_is_consultant', 'clinician_signatory_title',
    )
    list_display = ('id', 'nhs_number', 'pt_first_name', 'pt_last_name')
    list_display_links = ('id', 'nhs_number')
    search_fields = ('nhs_number', 'pt_first_name', 'pt_last_name')


# -----------------------------------------------------------------------------
# Patient lookup
# -----------------------------------------------------------------------------

class PatientLookupDevAdmin(ReadOnlyModelAdmin):
    readonly_fields = (
        # Lookup details
        'lookup_at', 'source_db', 'nhs_number',

        # Patient
        'pt_found',
        'pt_local_id_description', 'pt_local_id_number',
        'pt_dob', 'pt_dod', 'pt_dead', 'pt_discharged', 'pt_sex',
        'pt_title', 'pt_first_name', 'pt_last_name',
        'pt_address_1', 'pt_address_2', 'pt_address_3', 'pt_address_4',
        'pt_address_5', 'pt_address_6', 'pt_address_7',
        'pt_telephone', 'pt_email',

        # GP
        'gp_found',
        'gp_title', 'gp_first_name', 'gp_last_name',
        'gp_address_1', 'gp_address_2', 'gp_address_3', 'gp_address_4',
        'gp_address_5', 'gp_address_6', 'gp_address_7',
        'gp_telephone', 'gp_email',

        # Clinician
        'clinician_found',
        'clinician_title', 'clinician_first_name', 'clinician_last_name',
        'clinician_address_1', 'clinician_address_2', 'clinician_address_3',
        'clinician_address_4', 'clinician_address_5', 'clinician_address_6',
        'clinician_address_7',
        'clinician_telephone', 'clinician_email',
        'clinician_is_consultant', 'clinician_signatory_title',

        # Decisions
        'decisions', 'secret_decisions',

        # Extras
        'get_test_views',
    )
    fields = readonly_fields
    date_hierarchy = 'lookup_at'
    list_display = ('id', 'nhs_number',
                    'pt_first_name', 'pt_last_name', 'pt_dob')
    search_fields = ('nhs_number', 'pt_first_name', 'pt_last_name')

    def get_test_views(self, obj: PatientLookup) -> str:
        return '''
            <a href="{}">Draft letter to patient re first traffic-light
                choice (as HTML)</a><br>
            <a href="{}">Draft letter to patient re first traffic-light
                choice (as PDF)</a>
        '''.format(
            reverse('draft_first_traffic_light_letter', args=[obj.id, "html"]),
            reverse('draft_first_traffic_light_letter', args=[obj.id, "pdf"]),
        )
    get_test_views.short_description = "Test views"
    get_test_views.allow_tags = True


# -----------------------------------------------------------------------------
# Consent mode
# -----------------------------------------------------------------------------

class ConsentModeInline(admin.TabularInline):
    model = ConsentMode


class ConsentModeAdminForm(forms.ModelForm):
    def clean(self) -> Dict[str, Any]:
        if not self.cleaned_data.get('changed_by_clinician_override'):
            kwargs = {}
            for field in Decision.FIELDS:
                kwargs[field] = self.cleaned_data.get(field)
            decision = Decision(**kwargs)
            decision.validate_decision()
        return self.cleaned_data


class ConsentModeMgrAdmin(AddOnlyModelAdmin):
    # To switch off the Boolean icons: replace exclude_entirely with
    # exclude_entirely_col in the fieldlist, and define the function as:
    #
    # def exclude_entirely_col(self, obj):
    #     return obj.exclude_entirely
    # exclude_entirely_col.boolean = False
    #
    # Can use get_fields(self, request, obj=None) and get_readonly_fields(...)
    # to customize icon behaviour depending on whether we're adding or
    # editing.

    form = ConsentModeAdminForm
    fields = [
        'nhs_number',
        'exclude_entirely',
        'consent_mode',
        # 'consent_after_discharge',
        # 'max_approaches_per_year',
        # 'other_requests',
        'prefers_email',
        'changed_by_clinician_override',
    ] + Decision.FIELDS
    list_display = (
        'id', 'nhs_number', 'consent_mode',
        # 'consent_after_discharge'
        'source',
    )
    list_display_links = ('id', 'nhs_number')
    search_fields = ('nhs_number', )
    list_filter = (
        'consent_mode',
        # 'consent_after_discharge',
        'exclude_entirely',
        'prefers_email'
    )
    date_hierarchy = 'created_at'

    fields_for_viewing = replace_in_list(fields, {
        'exclude_entirely': 'exclude_entirely2',
        # 'consent_after_discharge': 'consent_after_discharge2',
        'prefers_email': 'prefers_email2',
        'changed_by_clinician_override': 'changed_by_clinician_override2',
    })
    exclude_entirely2 = disable_bool_icon('exclude_entirely', ConsentMode)
    consent_after_discharge2 = disable_bool_icon(
        'consent_after_discharge', ConsentMode)
    prefers_email2 = disable_bool_icon('prefers_email', ConsentMode)
    changed_by_clinician_override2 = disable_bool_icon(
        'changed_by_clinician_override', ConsentMode)

    # Populate the created_by field automatically, with the two functions below
    # https://code.djangoproject.com/wiki/CookBookNewformsAdminAndUser
    def save_model(self,
                   request: HttpRequest,
                   obj: ConsentMode,
                   form: forms.ModelForm,
                   change: bool) -> None:
        obj.current = False  # NOT YET; set by process_change()
        obj.created_by = request.user
        obj.save()
        transaction.on_commit(
            lambda: process_consent_change.delay(obj.id)
        )  # Asynchronous
        # Without transaction.on_commit, we get a RACE CONDITION:
        # object is received in the pre-save() state.
        self.message_user(
            request,
            "Consent mode will be changed. You will be e-mailed regarding "
            "the letter to patient (+/- re withdrawal of consent to "
            "researchers).")

    # def save_formset(self, request, form, formset, change):
    #     if formset.model == ConsentMode:
    #         instances = formset.save(commit=False)
    #         for instance in instances:
    #             instance.created_by = request.user
    #             instance.save()
    #     else:
    #         formset.save()

    # Restrict to current ones
    def get_queryset(self, request: HttpRequest) -> QuerySet:
        qs = super().get_queryset(request)
        return qs.filter(current=True)


class ConsentModeDevAdmin(ReadOnlyModelAdmin):
    readonly_fields = [
        'nhs_number',
        'current', 'created_at', 'created_by',
        'exclude_entirely', 'consent_mode', 'consent_after_discharge',
        'max_approaches_per_year', 'other_requests', 'prefers_email',
        'changed_by_clinician_override',
        'source',
    ] + Decision.FIELDS + [
        'get_test_views',
    ]
    fields = readonly_fields
    list_display = (
        'id',
        'current',
        'nhs_number',
        'consent_mode',
        # 'consent_after_discharge'
        'source',
    )
    search_fields = ('nhs_number', )
    list_filter = (
        'current',
        'consent_mode',
        # 'consent_after_discharge',
        'exclude_entirely',
        'prefers_email'
    )

    def get_test_views(self, obj: ConsentMode) -> str:
        return '''
            <a href="{}">Draft letter to patient confirming traffic-light
                choice (as HTML)</a><br>
            <a href="{}">Draft letter to patient confirming traffic-light
                choice (as PDF)</a>
        '''.format(
            reverse('draft_confirm_traffic_light_letter',
                    args=[obj.id, "html"]),
            reverse('draft_confirm_traffic_light_letter',
                    args=[obj.id, "pdf"]),
        )
    get_test_views.short_description = "Test views"
    get_test_views.allow_tags = True


# -----------------------------------------------------------------------------
# Team rep
# -----------------------------------------------------------------------------

class TeamRepMgrAdmin(admin.ModelAdmin):
    fields = ('team', 'user')
    list_display = ('team', 'user')
    search_fields = ('team', )


# -----------------------------------------------------------------------------
# Charity payments
# -----------------------------------------------------------------------------

class CharityPaymentRecordMgrAdmin(AddOnlyModelAdmin):
    fields = ('payee', 'amount')
    fields_for_viewing = fields
    list_display = ('id', 'created_at', 'payee', 'amount')
    list_display_links = ('id', 'created_at')
    search_fields = ('payee', )
    date_hierarchy = 'created_at'


# -----------------------------------------------------------------------------
# Contact request
# -----------------------------------------------------------------------------

class ClinicianRespondedListFilter(SimpleListFilter):
    title = 'clinician responded'
    parameter_name = 'clinician_responded'

    def lookups(self,
                request: HttpRequest,
                model_admin: admin.ModelAdmin) -> Iterable[Tuple[str, str]]:
        return (
            ('y', "Clinician responded"),
            ('n', "Clinician asked but hasn’t responded"),
        )

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        if self.value() == 'y':
            return (
                queryset
                .filter(decided_send_to_clinician=True)
                .filter(clinician_response__responded=True)
            )
        if self.value() == 'n':
            return (
                queryset
                .filter(decided_send_to_clinician=True)
                .filter(clinician_response__responded=False)
            )


class ContactRequestMgrAdmin(ReadOnlyModelAdmin):
    NONCONFIDENTIAL_FIELDS = (
        'id', 'created_at', 'request_by', 'get_study',
        'request_direct_approach',
        'lookup_nhs_number', 'lookup_rid', 'lookup_mrid',
        'processed',
        'get_consent_mode',
        'approaches_in_past_year',
        'decisions',
        'decided_no_action',
        'decided_send_to_researcher',
        'decided_send_to_clinician',
        'clinician_involvement',
        'consent_withdrawn',
        'consent_withdrawn_at',
        'get_letters', 'get_emails',
    )
    fields = NONCONFIDENTIAL_FIELDS + ('get_clinician_email_address',
                                       'get_clinician_responded')
    readonly_fields = fields
    NONCONFIDENTIAL_LIST_DISPLAY = (
        'id', 'created_at', 'request_by', 'study',
        'lookup_nhs_number', 'lookup_rid', 'lookup_mrid',
        'decided_no_action', 'decided_send_to_researcher',
        'decided_send_to_clinician', 'get_clinician_responded',
    )
    list_display = NONCONFIDENTIAL_LIST_DISPLAY + (
        'get_clinician_email_address',
    )
    list_filter = ('decided_no_action', 'decided_send_to_researcher',
                   'decided_send_to_clinician', ClinicianRespondedListFilter)
    list_select_related = (
        'clinician_response',
        'request_by',
        'study__lead_researcher',
        'patient_lookup',
        'consent_mode',
    )
    date_hierarchy = 'created_at'

    def get_consent_mode(self, obj: ContactRequest) -> ConsentMode:
        consent_mode = obj.consent_mode
        return consent_mode.consent_mode
    get_consent_mode.short_description = "Consent mode"

    def get_study(self, obj: ContactRequest) -> str:
        return admin_view_fk_link(self, obj, "study")
    get_study.allow_tags = True
    get_study.short_description = "Study"

    def get_clinician_email_address(self, obj: ContactRequest) -> str:
        if obj.decided_send_to_clinician:
            return obj.patient_lookup.clinician_email
        else:
            return ''
    get_clinician_email_address.short_description = "Clinician e-mail address"

    def get_clinician_responded(self, obj: ContactRequest) -> bool:
        if not hasattr(obj, 'clinician_response'):
            return False
        return obj.clinician_response.responded
    get_clinician_responded.short_description = "Clinician responded"
    get_clinician_responded.boolean = True

    def get_letters(self, obj: ContactRequest) -> str:
        return admin_view_reverse_fk_links(self, obj, "letter_set")
    get_letters.short_description = "Letter(s)"
    get_letters.allow_tags = True

    def get_emails(self, obj: ContactRequest) -> str:
        return admin_view_reverse_fk_links(self, obj, "email_set")
    get_emails.short_description = "E-mail(s)"
    get_emails.allow_tags = True


class ContactRequestResAdmin(ContactRequestMgrAdmin):
    fields = ContactRequestMgrAdmin.NONCONFIDENTIAL_FIELDS
    readonly_fields = fields
    list_display = ContactRequestMgrAdmin.NONCONFIDENTIAL_LIST_DISPLAY

    def get_queryset(self, request: HttpRequest) -> QuerySet:
        qs = super().get_queryset(request)
        studies = Study.filter_studies_for_researcher(Study.objects.all(),
                                                      request.user)
        return qs.filter(study__in=studies)

    def has_module_permission(self, request: HttpRequest) -> bool:
        return request.user.is_staff

    def has_change_permission(self,
                              request: HttpRequest,
                              obj: ContactRequest = None) -> bool:
        return request.user.is_staff


class ContactRequestDevAdmin(ContactRequestMgrAdmin):
    fields = ContactRequestMgrAdmin.NONCONFIDENTIAL_FIELDS + (
        # RDBM can also see
        'get_clinician_email_address',
        'get_link_clinician_email',
        'get_clinician_responded',
        'get_link_clinician_response',

        # properly secret
        'nhs_number',
        'get_patient_lookup',

        # exploratory test views
        'get_test_views',
    )
    readonly_fields = fields
    list_display = (
        'id', 'created_at', 'request_by', 'study',
        'nhs_number',
        'decided_no_action', 'decided_send_to_researcher',
        'decided_send_to_clinician', 'get_clinician_responded',
        'get_clinician_email_address',
    )

    def get_link_clinician_email(self, obj: ContactRequest) -> str:
        return admin_view_reverse_fk_links(self, obj, "email_set")
    get_link_clinician_email.short_description = "E-mail to clinician"
    get_link_clinician_email.allow_tags = True

    def get_link_clinician_response(self, obj: ContactRequest) -> str:
        return admin_view_fk_link(self, obj, "clinician_response")
    get_link_clinician_response.short_description = "Clinician response"
    get_link_clinician_response.allow_tags = True

    def get_patient_lookup(self, obj: ContactRequest) -> str:
        return admin_view_fk_link(self, obj, "patient_lookup")
    get_patient_lookup.short_description = "Patient lookup"
    get_patient_lookup.allow_tags = True

    def get_consent_mode(self, obj: ContactRequest) -> str:
        return admin_view_fk_link(self, obj, "consent_mode")
    get_consent_mode.short_description = "Consent mode"
    get_consent_mode.allow_tags = True

    def get_letters(self, obj: ContactRequest) -> str:
        return admin_view_reverse_fk_links(self, obj, "letter_set")
    get_letters.allow_tags = True
    get_letters.short_description = "Letter(s)"

    def get_test_views(self, obj: ContactRequest) -> str:
        return '''
            <a href="{}">Draft e-mail to clinician</a><br>
            <a href="{}">Draft letter from clinician to patient re study
                (HTML)</a></br>
            <a href="{}">Draft letter from clinician to patient re study
                (PDF)</a></br>
            <a href="{}">Decision form to patient re study (HTML)</a><br>
            <a href="{}">Decision form to patient re study (PDF)</a><br>
            <a href="{}">Draft approval letter to researcher (HTML)</a><br>
            <a href="{}">Draft approval letter to researcher (PDF)</a><br>
            <a href="{}">Draft approval covering e-mail to researcher</a><br>
            <a href="{}">Draft withdrawal letter to researcher (HTML)</a><br>
            <a href="{}">Draft withdrawal letter to researcher (PDF)</a><br>
            <a href="{}">Draft withdrawal covering e-mail to researcher</a>
        '''.format(
            reverse('draft_clinician_email', args=[obj.id]),
            reverse('draft_letter_clinician_to_pt_re_study',
                    args=[obj.id, "html"]),
            reverse('draft_letter_clinician_to_pt_re_study',
                    args=[obj.id, "pdf"]),
            reverse('decision_form_to_pt_re_study',
                    args=[obj.id, "html"]),
            reverse('decision_form_to_pt_re_study',
                    args=[obj.id, "pdf"]),
            reverse('draft_approval_letter', args=[obj.id, "html"]),
            reverse('draft_approval_letter', args=[obj.id, "pdf"]),
            reverse('draft_approval_email', args=[obj.id]),
            reverse('draft_withdrawal_letter', args=[obj.id, "html"]),
            reverse('draft_withdrawal_letter', args=[obj.id, "pdf"]),
            reverse('draft_withdrawal_email', args=[obj.id]),
        )
    get_test_views.short_description = "Test views"
    get_test_views.allow_tags = True


# -----------------------------------------------------------------------------
# Clinician response
# -----------------------------------------------------------------------------

class ClinicianResponseDevAdmin(ReadOnlyModelAdmin):
    fields = [
        'created_at', 'contact_request', 'token',
        'responded', 'responded_at', 'response_route',
        'email_choice', 'response',
        'veto_reason', 'ineligible_reason', 'pt_uncontactable_reason',
        'clinician_confirm_name',
        'charity_amount_due',

        'get_contact_request',
    ]
    readonly_fields = fields
    date_hierarchy = 'created_at'

    def get_contact_request(self, obj: ClinicianResponse) -> str:
        return admin_view_fk_link(self, obj, "contact_request")
    get_contact_request.allow_tags = True
    get_contact_request.short_description = "Contact request"


# -----------------------------------------------------------------------------
# Patient response
# -----------------------------------------------------------------------------

class PatientResponseAdminForm(forms.ModelForm):
    def clean(self) -> Dict[str, Any]:
        kwargs = {}
        for field in Decision.FIELDS:
            kwargs[field] = self.cleaned_data.get(field)
        decision = Decision(**kwargs)
        decision.validate_decision()
        return self.cleaned_data


class PatientResponseMgrAdmin(EditOnceOnlyModelAdmin):
    form = PatientResponseAdminForm
    fields = [
        'id', 'created_at', 'get_contact_request', 'response'
    ] + Decision.FIELDS
    readonly_fields = ['id', 'created_at', 'get_contact_request']
    date_hierarchy = 'created_at'

    # Populate the created_by field automatically, with the two functions below
    def save_model(self,
                   request: HttpRequest,
                   obj: PatientResponse,
                   form: forms.ModelForm,
                   change: bool) -> None:
        obj.recorded_by = request.user
        obj.save()
        # log.debug("PatientResponse: {}".format(modelrepr(obj)))
        transaction.on_commit(
            lambda: process_patient_response.delay(obj.id)
        )  # Asynchronous
        if obj.response == PatientResponse.YES:
            self.message_user(
                request,
                "Approval to researcher will be generated. You will be "
                "e-mailed if the system can't send it to the researcher.")

    def has_change_permission(self,
                              request: HttpRequest,
                              obj: PatientResponse = None) -> bool:
        if obj and obj.response:
            return False  # already saved
        return True

    def get_contact_request(self, obj: PatientResponse) -> str:
        return admin_view_fk_link(self, obj, "contact_request")
    get_contact_request.short_description = "Contact request"
    get_contact_request.allow_tags = True

    def get_queryset(self, request: HttpRequest) -> QuerySet:
        # Restrict to unresponded ones
        return super().get_queryset(request).filter(response__isnull=True)


class PatientResponseDevAdmin(ReadOnlyModelAdmin):
    fields = PatientResponseMgrAdmin.fields
    readonly_fields = fields
    date_hierarchy = 'created_at'

    def get_contact_request(self, obj: PatientResponse) -> str:
        return admin_view_fk_link(self, obj, "contact_request")
    get_contact_request.short_description = "Contact request"
    get_contact_request.allow_tags = True


# -----------------------------------------------------------------------------
# Letters
# -----------------------------------------------------------------------------

class LetterSendingStatusFilter(SimpleListFilter):
    title = "sending status"
    parameter_name = 'sending_status'

    def lookups(self,
                request: HttpRequest,
                model_admin: admin.ModelAdmin) -> Iterable[Tuple[str, str]]:
        return (
            ('sent_manually', "Sent manually"),
            ('not_sent_manually', "Not sent manually"),
            ('sent_by_email', "Sent by e-mail"),
            ('not_sent_by_email', "Not sent by e-mail"),
            ('require_sending', "REQUIRE SENDING"),
        )

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        if self.value() == 'sent_manually':
            return queryset.filter(sent_manually_at__isnull=False)
        if self.value() == 'not_sent_manually':
            return queryset.filter(sent_manually_at__isnull=True)
        if self.value() == 'sent_by_email':
            return (
                queryset
                .filter(email__emailtransmission__sent=True)
                .distinct()
            )
        if self.value() == 'not_sent_by_email':
            return queryset.exclude(email__emailtransmission__sent=True)
        if self.value() == 'require_sending':
            return (
                queryset
                # Restrict to letters not sent manually
                .filter(sent_manually_at__isnull=True)
                # Exclude letters successfully sent by e-mail
                .exclude(email__emailtransmission__sent=True)
            )


class LetterDevAdmin(ReadOnlyModelAdmin):
    fields = ('id', 'created_at', 'pdf',
              'to_clinician', 'to_researcher', 'to_patient',
              'get_study', 'get_contact_request',
              'sent_manually_at', 'get_emails')
    readonly_fields = fields
    list_display = ('id', 'created_at',
                    'to_clinician', 'to_researcher', 'to_patient',
                    'study', 'contact_request',
                    'sent_manually_at')
    list_filter = (LetterSendingStatusFilter,
                   'to_clinician', 'to_researcher', 'to_patient')
    list_select_related = (
        'study__lead_researcher',
        'contact_request',
    )
    date_hierarchy = 'created_at'
    # ... see also http://stackoverflow.com/questions/991926/custom-filter-in-django-admin-on-django-1-3-or-below  # noqa
    actions = ['mark_sent']

    def mark_sent(self, request: HttpRequest, queryset: QuerySet) -> None:
        ids = []
        for letter in queryset:
            letter.mark_sent()
            ids.append(letter.id)
        self.message_user(
            request,
            "{} letter(s) were marked as sent: IDs {}.".format(len(ids),
                                                               str(ids)))
    mark_sent.short_description = "Mark selected letters as printed/sent"

    def get_study(self, obj: Letter) -> str:
        return admin_view_fk_link(self, obj, "study")
    get_study.allow_tags = True
    get_study.short_description = "Study"

    def get_contact_request(self, obj: Letter) -> str:
        return admin_view_fk_link(self, obj, "contact_request")
    get_contact_request.allow_tags = True
    get_contact_request.short_description = "Contact request"

    def get_emails(self, obj: Letter) -> str:
        return admin_view_reverse_fk_links(self, obj, "email_set")
    get_emails.allow_tags = True
    get_emails.short_description = "E-mail(s)"


class LetterMgrAdmin(LetterDevAdmin):
    """Restrict to letters visible to a researcher."""
    def get_queryset(self, request: HttpRequest) -> QuerySet:
        return (
            super().get_queryset(request)
            .filter(Q(to_researcher=True) | Q(rdbm_may_view=True))
        )


class LetterResAdmin(LetterDevAdmin):
    fields = ('id', 'created_at', 'get_pdf',
              'to_clinician', 'to_researcher', 'to_patient',
              'study', 'contact_request',
              'sent_manually_at', 'email')
    readonly_fields = fields
    """Restrict to letters visible to a researcher."""
    def get_queryset(self, request: HttpRequest) -> QuerySet:
        qs = super().get_queryset(request).filter(to_researcher=True)
        studies = Study.filter_studies_for_researcher(Study.objects.all(),
                                                      request.user)
        return qs.filter(study__in=studies)

    def has_module_permission(self, request: HttpRequest) -> bool:
        return request.user.is_staff

    def has_change_permission(self,
                              request: HttpRequest,
                              obj: Letter = None) -> bool:
        return request.user.is_staff

    def get_pdf(self, obj: Letter) -> str:
        if not obj.pdf:
            return "(Missing)"
        return '<a href="{}">PDF</a>'.format(
            reverse("letter", args=[obj.id]))
    get_pdf.short_description = "Letter PDF"
    get_pdf.allow_tags = True


# =============================================================================
# User profiles
# =============================================================================

class UserProfileInline(admin.StackedInline):
    model = UserProfile
    max_num = 1
    can_delete = False
    inlines = [StudyInline]
    fields = ('per_page', 'line_length',
              'collapse_at_len', 'collapse_at_n_lines',
              'is_developer',
              'title',
              'telephone',
              'address_1', 'address_2', 'address_3', 'address_4',
              'address_5', 'address_6', 'address_7',
              'get_studies_as_lead', 'get_studies_as_researcher',
              'enough_info_for_researcher')
    readonly_fields = ('get_studies_as_lead', 'get_studies_as_researcher',
                       'enough_info_for_researcher')

    def get_studies_as_lead(self, obj: settings.AUTH_USER_MODEL) -> str:
        studies = obj.user.studies_as_lead.all()
        return render_to_string('shortlist_studies.html', {'studies': studies})
    get_studies_as_lead.short_description = "Studies as lead researcher"
    get_studies_as_lead.allow_tags = True

    def get_studies_as_researcher(self, obj: settings.AUTH_USER_MODEL) -> str:
        studies = obj.user.studies_as_researcher.all()
        return render_to_string('shortlist_studies.html', {'studies': studies})
    get_studies_as_researcher.short_description = "Studies as researcher"
    get_studies_as_researcher.allow_tags = True

    def enough_info_for_researcher(self,
                                   obj: settings.AUTH_USER_MODEL) -> bool:
        return (
            bool(obj.title) and
            bool(obj.user.first_name) and
            bool(obj.user.last_name)
        )
    enough_info_for_researcher.short_description = (
        "Enough info for researcher status (title, firstname, lastname)?"
    )
    enough_info_for_researcher.boolean = True


class ExtendedUserMgrAdmin(UserAdmin):
    inlines = [UserProfileInline]
    list_display = (
        'username', 'email', 'first_name', 'last_name', 'is_staff',
        'enough_info_for_researcher',
    )

    def enough_info_for_researcher(self,
                                   obj: settings.AUTH_USER_MODEL) -> bool:
        return (
            bool(obj.profile.title) and
            bool(obj.first_name) and
            bool(obj.last_name)
        )
    enough_info_for_researcher.short_description = (
        "Enough researcher info?"
    )
    enough_info_for_researcher.boolean = True


# =============================================================================
# Assemble main admin site
# =============================================================================
# http://stackoverflow.com/questions/4938491/django-admin-change-header-django-administration-text  # noqa
# http://stackoverflow.com/questions/3400641/how-do-i-inline-edit-a-django-user-profile-in-the-admin-interface  # noqa

class MgrAdminSite(admin.AdminSite):
    # Text to put at the end of each page's <title>.
    site_title = ugettext_lazy(settings.RESEARCH_DB_TITLE + ' manager admin')
    # Text to put in each page's <h1>.
    site_header = ugettext_lazy(settings.RESEARCH_DB_TITLE + ": manager admin")
    # URL for the "View site" link at the top of each admin page.
    site_url = settings.FORCE_SCRIPT_NAME + "/"
    # Text to put at the top of the admin index page.
    index_title = ugettext_lazy(settings.RESEARCH_DB_TITLE +
                                ' site administration for RDBM')
    index_template = 'admin/viewchange_admin_index.html'
    app_index_template = 'admin/viewchange_admin_app_index.html'


mgr_admin_site = MgrAdminSite(name="mgradmin")
mgr_admin_site.disable_action('delete_selected')
# ... particularly for e-mail where we manually specify a bulk action (resend)
# https://docs.djangoproject.com/en/1.8/ref/contrib/admin/actions/
mgr_admin_site.register(CharityPaymentRecord, CharityPaymentRecordMgrAdmin)
mgr_admin_site.register(ConsentMode, ConsentModeMgrAdmin)
mgr_admin_site.register(ContactRequest, ContactRequestMgrAdmin)
mgr_admin_site.register(Email, EmailMgrAdmin)
mgr_admin_site.register(Leaflet, LeafletMgrAdmin)
mgr_admin_site.register(Letter, LetterMgrAdmin)
mgr_admin_site.register(PatientResponse, PatientResponseMgrAdmin)
mgr_admin_site.register(QueryAudit, QueryMgrAdmin)
mgr_admin_site.register(Study, StudyMgrAdmin)
mgr_admin_site.register(TeamRep, TeamRepMgrAdmin)
mgr_admin_site.register(User, ExtendedUserMgrAdmin)


# =============================================================================
# Assemble secondary (developer) admin site
# =============================================================================
# http://stackoverflow.com/questions/4938491/django-admin-change-header-django-administration-text  # noqa
# http://stackoverflow.com/questions/3400641/how-do-i-inline-edit-a-django-user-profile-in-the-admin-interface  # noqa

class DevAdminSite(admin.AdminSite):
    site_title = ugettext_lazy(settings.RESEARCH_DB_TITLE + ' dev admin')
    site_header = ugettext_lazy(settings.RESEARCH_DB_TITLE +
                                ": developer admin")
    site_url = settings.FORCE_SCRIPT_NAME + "/"
    index_title = ugettext_lazy(settings.RESEARCH_DB_TITLE +
                                ' developer administration')
    index_template = 'admin/viewchange_admin_index.html'
    app_index_template = 'admin/viewchange_admin_app_index.html'


dev_admin_site = DevAdminSite(name="devadmin")
dev_admin_site.disable_action('delete_selected')
# Where no specific DevAdmin version exists, use the MgrAdmin
dev_admin_site.register(CharityPaymentRecord, CharityPaymentRecordMgrAdmin)
dev_admin_site.register(ClinicianResponse, ClinicianResponseDevAdmin)
dev_admin_site.register(ConsentMode, ConsentModeDevAdmin)
dev_admin_site.register(ContactRequest, ContactRequestDevAdmin)
dev_admin_site.register(DummyPatientSourceInfo, DummyPatientSourceInfoDevAdmin)
dev_admin_site.register(Email, EmailDevAdmin)
dev_admin_site.register(Leaflet, LeafletMgrAdmin)
dev_admin_site.register(Letter, LetterDevAdmin)
dev_admin_site.register(PatientLookup, PatientLookupDevAdmin)
dev_admin_site.register(PatientResponse, PatientResponseDevAdmin)
dev_admin_site.register(QueryAudit, QueryMgrAdmin)
dev_admin_site.register(Study, StudyMgrAdmin)
dev_admin_site.register(TeamRep, TeamRepMgrAdmin)
dev_admin_site.register(User, ExtendedUserMgrAdmin)


# =============================================================================
# Assemble tertiary (researcher) admin site
# =============================================================================

class ResearcherAdminSite(admin.AdminSite):
    site_title = ugettext_lazy(settings.RESEARCH_DB_TITLE +
                               ' researcher admin views')
    site_header = ugettext_lazy(settings.RESEARCH_DB_TITLE +
                                ": researcher admin")
    site_url = settings.FORCE_SCRIPT_NAME + "/"
    index_title = ugettext_lazy("View/manage your studies")
    index_template = 'admin/viewchange_admin_index.html'
    app_index_template = 'admin/viewchange_admin_app_index.html'


res_admin_site = ResearcherAdminSite(name="resadmin")
res_admin_site.disable_action('delete_selected')
res_admin_site.register(Study, StudyResAdmin)
res_admin_site.register(Leaflet, LeafletResAdmin)
res_admin_site.register(Email, EmailResAdmin)
res_admin_site.register(Letter, LetterResAdmin)
res_admin_site.register(ContactRequest, ContactRequestResAdmin)

"""
Problem with non-superusers not seeing any apps:
- http://stackoverflow.com/questions/1929707/django-admin-not-seeing-any-app-permission-problem  # noqa
  ... but django.contrib.auth.backends.ModelBackend won't load in INSTALLED_APPS  # noqa
- log.debug("registered: {}".format(res_admin_site.is_registered(Leaflet)))
  ... OK
  ... and anyway, it works for superusers
- app_list is blank in the template; this is set in AdminSite.index()
  (in django/contrib/admin/sites.py)
  So the failure is either in
        model_admin.has_module_permission(request)
            return request.user.has_module_perms(self.opts.app_label)
    or  model_admin.get_model_perms(request)
  They're in django/contrib/admin/options.py; ModelAdmin and BaseModelAdmin

  From a Werkzeug console in home view:

    from core.admin import LeafletResAdmin
    LeafletResAdmin.has_module_permission(request)
    # ... fails (class not instance!) but shows code, so:
    request.user.has_module_perms('resadmin')
    # ... False - so HERE'S a problem.

  Solution: add these to relevant ModelAdmin classes:

    def has_module_permission(self, request):
        return request.user.is_staff

    def has_change_permission(self, request, obj=None):
        return request.user.is_staff
"""



#!/usr/bin/env python
# crate_anon/tools/generate_new_django_secret_key.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

See django.core.management.commands.startproject.Command.handle
"""


from django.utils.crypto import get_random_string


def main():
    chars = 'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)'
    key = get_random_string(50, chars)
    print(key)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# crate_anon/common/contenttypes.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Many of these can be extracted:

import mimetypes
mimetypes.types_map['.zip']  # application/zip -- this is built in
mimetypes.types_map['.xlsx']  # fails
mimetypes.init()
mimetypes.types_map['.xlsx']  # application/vnd.openxmlformats-officedocument.spreadsheetml.sheet
# ... must read some local thing...

"""  # noqa

CONTENTTYPE_PDF = "application/pdf"

CONTENTTYPE_ZIP = "application/zip"
# ... http://stackoverflow.com/questions/4411757/zip-mime-types-when-to-pick-which-one  # noqa

CONTENTTYPE_TSV = "text/tab-separated-values"
CONTENTTYPE_CSV = "text/csv"
# http://stackoverflow.com/questions/264256/what-is-the-best-mime-type-and-extension-to-use-when-exporting-tab-delimited  # noqa
# http://www.iana.org/assignments/media-types/text/tab-separated-values

CONTENTTYPE_XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"  # noqa



#!/usr/bin/env python
# crate_anon/anonymise/launch_multiprocess_anonymiser.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Launch anonymise.py in multiprocess mode.

Previous bash version: see

    http://stackoverflow.com/questions/356100
    http://stackoverflow.com/questions/1644856
    http://stackoverflow.com/questions/8903239
    http://stackoverflow.com/questions/1951506
    Note: $! is the process ID of last process launched in background
    http://stackoverflow.com/questions/59895

Python version: see

    http://stackoverflow.com/questions/23611396/python-execute-cat-subprocess-in-parallel  # noqa
    http://stackoverflow.com/questions/320232/ensuring-subprocesses-are-dead-on-exiting-python-program  # noqa
    http://stackoverflow.com/questions/641420/how-should-i-log-while-using-multiprocessing-in-python  # noqa
"""

import argparse
import logging
import multiprocessing
import sys
import time

from crate_anon.common.logsupport import configure_logger_for_colour
from crate_anon.common.subproc import (
    check_call_process,
    run_multiple_processes,
)
from crate_anon.version import VERSION, VERSION_DATE

log = logging.getLogger(__name__)

ANONYMISER = 'crate_anon.anonymise.anonymise_cli'

CPUCOUNT = multiprocessing.cpu_count()


# =============================================================================
# Main
# =============================================================================

def main() -> None:
    version = "Version {} ({})".format(VERSION, VERSION_DATE)
    description = (
        "Runs the CRATE anonymiser in parallel. {}. Note that all arguments "
        "not specified here are passed to the underlying script "
        "(see crate_anonymise --help).".format(version))
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "--nproc", "-n", nargs="?", type=int, default=CPUCOUNT,
        help="Number of processes (default on this "
             "machine: {})".format(CPUCOUNT))
    parser.add_argument(
        '--verbose', '-v', action='store_true',
        help="Be verbose")
    args, unknownargs = parser.parse_known_args()

    loglevel = logging.DEBUG if args.verbose else logging.INFO
    rootlogger = logging.getLogger()
    configure_logger_for_colour(rootlogger, level=loglevel)

    common_options = ["-v"] * (1 if args.verbose else 0) + unknownargs

    log.debug("common_options: {}".format(common_options))

    nprocesses_patient = args.nproc
    nprocesses_nonpatient = args.nproc
    nprocesses_index = args.nproc

    # -------------------------------------------------------------------------
    # Setup
    # -------------------------------------------------------------------------

    # Start.
    time_start = time.time()

    # -------------------------------------------------------------------------
    # Clean/build the tables. Only run one copy of this!
    # -------------------------------------------------------------------------
    # CALL USING "python -m my.module"; DO NOT run the script as an executable.
    # If you run a Python script as an executable, it gets added to the
    # PYTHONPATH. Then, when your script says "import regex" (meaning the
    # system module), it might import "regex.py" from the same directory (which
    # it wouldn't normally do, because Python 3 uses absolute not relative
    # imports).
    procargs = [
        sys.executable, '-m', ANONYMISER,
        '--dropremake', '--processcluster=STRUCTURE'
    ] + common_options
    check_call_process(procargs)

    # -------------------------------------------------------------------------
    # Build opt-out lists. Only run one copy of this!
    # -------------------------------------------------------------------------
    procargs = [
        sys.executable, '-m', ANONYMISER,
        '--optout', '--processcluster=OPTOUT',
        '--skip_dd_check'
    ] + common_options
    check_call_process(procargs)

    # -------------------------------------------------------------------------
    # Now run lots of things simultaneously:
    # -------------------------------------------------------------------------
    # It'd be less confusing if we have a single numbering system across all,
    # rather than numbering separately for patient and non-patient processes.
    # However, each group divides its work by its process number, so that
    # won't fly (for n processes it wants to see processes numbered from 0 to
    # n - 1 inclusive).

    # (a) patient tables
    args_list = []
    for procnum in range(nprocesses_patient):
        procargs = [
            sys.executable, '-m', ANONYMISER,
            '--patienttables',
            '--processcluster=PATIENT',
            '--nprocesses={}'.format(nprocesses_patient),
            '--process={}'.format(procnum),
            '--skip_dd_check'
        ] + common_options
        args_list.append(procargs)
    for procnum in range(nprocesses_nonpatient):
        procargs = [
            sys.executable, '-m', ANONYMISER,
            '--nonpatienttables',
            '--processcluster=NONPATIENT',
            '--nprocesses={}'.format(nprocesses_nonpatient),
            '--process={}'.format(procnum),
            '--skip_dd_check'
        ] + common_options
        args_list.append(procargs)
    run_multiple_processes(args_list)  # Wait for them all to finish

    time_middle = time.time()

    # -------------------------------------------------------------------------
    # Now do the indexing, if nothing else failed.
    # (Always fastest to index last.)
    # -------------------------------------------------------------------------
    args_list = [
        [
            sys.executable, '-m', ANONYMISER,
            '--index',
            '--processcluster=INDEX',
            '--nprocesses={}'.format(nprocesses_index),
            '--process={}'.format(procnum),
            '--skip_dd_check'
        ] + common_options for procnum in range(nprocesses_index)
    ]
    run_multiple_processes(args_list)

    # -------------------------------------------------------------------------
    # Finished.
    # -------------------------------------------------------------------------
    time_end = time.time()
    main_dur = time_middle - time_start
    index_dur = time_end - time_middle
    total_dur = time_end - time_start
    print("Time taken: main {} s, indexing {} s, total {} s".format(
        main_dur, index_dur, total_dur))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# crate_anon/crateweb/core/dbfunc.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from collections import OrderedDict
# import logging
# log = logging.getLogger(__name__)
from typing import Any, Dict, Generator, List, Optional


def get_fieldnames_from_cursor(cursor) -> List[str]:
    """
    Get fieldnames from an executed cursor.
    """
    return [i[0] for i in cursor.description]


def tsv_escape(x: Any) -> str:
    """
    Escape data for tab-separated value format.
    """
    if x is None:
        return ""
    x = str(x)
    return x.replace("\t", "\\t").replace("\n", "\\n")


def make_tsv_row(values: List[Any]) -> str:
    return "\t".join([tsv_escape(x) for x in values]) + "\n"


def genrows(cursor, arraysize: int = 1000) -> Generator[List[Any], None, None]:
    """Generate all rows from a cursor."""
    # http://code.activestate.com/recipes/137270-use-generators-for-fetching-large-db-record-sets/  # noqa
    while True:
        results = cursor.fetchmany(arraysize)
        if not results:
            break
        for result in results:
            yield result


def genfirstvalues(cursor, arraysize: int = 1000) -> Generator[Any, None, None]:
    """Generate the first value in each row."""
    return (row[0] for row in genrows(cursor, arraysize))


def fetchallfirstvalues(cursor) -> List[Any]:
    """Return a list of the first value in each row."""
    return [row[0] for row in cursor.fetchall()]


def gendicts(cursor, arraysize: int = 1000) -> Generator[Dict[str, Any],
                                                         None, None]:
    """Generate all rows from a cursor as a list of OrderedDicts."""
    columns = get_fieldnames_from_cursor(cursor)
    return (
        OrderedDict(zip(columns, row))
        for row in genrows(cursor, arraysize)
    )


def dictfetchall(cursor) -> List[Dict[str, Any]]:
    """Return all rows from a cursor as a list of OrderedDicts."""
    columns = get_fieldnames_from_cursor(cursor)
    return [
        OrderedDict(zip(columns, row))
        for row in cursor.fetchall()
    ]


def dictfetchone(cursor) -> Optional[Dict[str, Any]]:
    """
    Return the next row from a cursor as an OrderedDict, or None
    """
    columns = get_fieldnames_from_cursor(cursor)
    row = cursor.fetchone()
    if not row:
        return None
    return OrderedDict(zip(columns, row))


def dictlist_to_tsv(dictlist: List[Dict[str, Any]]) -> str:
    if not dictlist:
        return ""
    fieldnames = dictlist[0].keys()
    tsv = "\t".join([tsv_escape(f) for f in fieldnames]) + "\n"
    for d in dictlist:
        tsv += "\t".join([tsv_escape(v) for v in d.values()]) + "\n"
    return tsv



#!/usr/bin/env python
# crate_anon/nlp_manager/regex_numbers.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""


# =============================================================================
# Mathematical operations and quantities
# =============================================================================

MULTIPLY = r"[x\*×⋅]"  # x, *, ×, ⋅
MULTIPLY_OR_SPACE = r"[x\*×⋅\s]"  # x, *, ×, ⋅, space
POWER = r"(?: \^ | \*\* )"  # ^, **
POWER_INC_E = r"(?: e | \^ | \*\* )"  # e, ^, **


def times_ten_to_power(n):
    return r"(?: {MULTIPLY}? \s* 10 \s* {POWER_INC_E} \s* {n})".format(
        MULTIPLY=MULTIPLY, POWER_INC_E=POWER_INC_E, n=n)


BILLION = times_ten_to_power(9)


# =============================================================================
# Number components
# =============================================================================
# Don't create components that are entirely optional; they're hard to test!

PLUS_SIGN = r"\+"  # don't forget to escape it
MINUS_SIGN = r"[-−–]"  # any of: ASCII hyphen-minus, Unicode minus, en dash
SIGN = r"(?: {plus} | {minus} )".format(plus=PLUS_SIGN, minus=MINUS_SIGN)

NO_MINUS_SIGN = r"(?!{minus})".format(minus=MINUS_SIGN)
# ... (?! something ) is a negative lookahead assertion

PLAIN_INTEGER = r"(?:\d+)"
# Numbers with commas: http://stackoverflow.com/questions/5917082
# ... then modified a little, because that fails with Python's regex module;
# (a) the "\d+" grabs things like "12,000" and thinks "aha, 12", so we have to
#     fix that by putting the "thousands" bit first; then
# (b) that has to be modified to contain at least one comma/thousands grouping
#     (or it will treat "9800" as "980").

PLAIN_INTEGER_W_THOUSAND_COMMAS = r"(?: (?: \d{1,3} (?:,\d{3})+ ) | \d+ )"
# ... plain integer allowing commas as a thousands separator
#       (1) a number with thousands separators, or
#       (2) a plain number
# ... NOTE: PUT THE ONE THAT NEEDS TO BE GREEDIER FIRST, i.e. the one with
# thousands separators

FLOATING_POINT_GROUP = r"(?: \. \d+ )"  # decimal point and further digits
SCIENTIFIC_NOTATION_EXPONENT = r"(?: E {sign}? \d+ )".format(
    sign=SIGN,  # optional
)
# ... Scientific notation does NOT offer non-integer exponents.
# Specifically, float("-3.4e-27") is fine, but float("-3.4e-27.1") isn't.


# =============================================================================
# Number types
# =============================================================================
# Beware of unsigned types. You may not want a sign, but if you use an
# unsigned type, "-3" will be read as "3".

UNSIGNED_INTEGER = PLAIN_INTEGER_W_THOUSAND_COMMAS
SIGNED_INTEGER = r"(?: {sign}? {integer} )".format(
    sign=SIGN,  # optional
    integer=PLAIN_INTEGER_W_THOUSAND_COMMAS,
)
UNSIGNED_FLOAT = r"(?: {nominus} {plus}? {integer} {fp}? )".format(
    nominus=NO_MINUS_SIGN,
    plus=PLUS_SIGN,  # optional
    integer=PLAIN_INTEGER_W_THOUSAND_COMMAS,
    fp=FLOATING_POINT_GROUP,  # optional
)
SIGNED_FLOAT = r"(?: {sign}? {integer} {fp}? )".format(
    sign=SIGN,  # optional
    integer=PLAIN_INTEGER_W_THOUSAND_COMMAS,
    fp=FLOATING_POINT_GROUP,  # optional
)
LIBERAL_NUMBER = r"(?: {sign}? {integer} {fp}? {exp}? )".format(
    sign=SIGN,  # optional
    integer=PLAIN_INTEGER_W_THOUSAND_COMMAS,
    fp=FLOATING_POINT_GROUP,  # optional
    exp=SCIENTIFIC_NOTATION_EXPONENT,  # optional
)



#!/usr/bin/env python
# crate_anon/preprocess/ddhint.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
from typing import Iterable, List, Tuple, Union

from sqlalchemy import MetaData
from sqlalchemy.engine import Engine

from crate_anon.common.lang import (
    get_case_insensitive_dict_key,
)
from crate_anon.common.sql import (
    add_indexes,
    drop_indexes,
)

log = logging.getLogger(__name__)


# =============================================================================
# DDHint class
# =============================================================================

class DDHint(object):
    def __init__(self) -> None:
        self._suppressed_tables = set()
        self._index_requests = {}  # dict of dicts

    def suppress_table(self, table: str) -> None:
        self._suppressed_tables.add(table)

    def suppress_tables(self, tables: Iterable[str]) -> None:
        for t in tables:
            self.suppress_table(t)

    def get_suppressed_tables(self) -> List[str]:
        return sorted(self._suppressed_tables)

    def add_source_index_request(self,
                                 table: str,
                                 columns: Union[str, Iterable[str]]) -> None:
        if isinstance(columns, str):
            columns = [columns]
        assert table, "Bad table: {}".format(repr(table))
        assert columns, "Bad columns: {}".format(repr(columns))
        index_name = 'crate_idx_' + '_'.join(columns)
        if table not in self._index_requests:
            self._index_requests[table] = {}
            if index_name not in self._index_requests[table]:
                self._index_requests[table][index_name] = {
                    'index_name': index_name,
                    'column': ', '.join(columns),
                    'unique': False,
                }

    def add_bulk_source_index_request(
            self,
            table_columns_list: Iterable[Tuple[str, Iterable[str]]]) -> None:
        for table, columns in table_columns_list:
            assert table, ("Bad table; table={}, table_columns_list={}".format(
                repr(table), repr(table_columns_list)))
            assert columns, (
                "Bad table; columns={}, table_columns_list={}".format(
                    repr(columns), repr(table_columns_list)))
            self.add_source_index_request(table, columns)

    def add_indexes(self, engine: Engine, metadata: MetaData) -> None:
        for tablename, tabledict in self._index_requests.items():
            indexdictlist = []
            for indexname, indexdict in tabledict.items():
                indexdictlist.append(indexdict)
            tablename_casematch = get_case_insensitive_dict_key(
                metadata.tables, tablename)
            if not tablename_casematch:
                log.warning("add_indexes: Skipping index as table {} "
                            "absent".format(tablename))
                continue
            table = metadata.tables[tablename_casematch]
            add_indexes(engine, table, indexdictlist)

    def drop_indexes(self, engine: Engine, metadata: MetaData) -> None:
        for tablename, tabledict in self._index_requests.items():
            index_names = list(tabledict.keys())
            tablename_casematch = get_case_insensitive_dict_key(
                metadata.tables, tablename)
            if not tablename_casematch:
                log.warning("add_indexes: Skipping index as table {} "
                            "absent".format(tablename))
                continue
            table = metadata.tables[tablename_casematch]
            drop_indexes(engine, table, index_names)



#!/usr/bin/env python
# crate_anon/common/fileops.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import fnmatch
import glob
import logging
import os
import shutil
from typing import List

log = logging.getLogger(__name__)


def mkdir_p(path: str) -> None:
    log.debug("mkdir_p: " + path)
    os.makedirs(path, exist_ok=True)


def copyglob(src: str, dest: str, allow_nothing: bool = False) -> None:
    something = False
    for filename in glob.glob(src):
        shutil.copy(filename, dest)
        something = True
    if something or allow_nothing:
        return
    raise ValueError("No files found matching: {}".format(src))


def copytree(src_dir: str, dest_parent: str) -> None:
    dirname = os.path.basename(os.path.normpath(src_dir))
    dest_dir = os.path.join(dest_parent, dirname)
    shutil.copytree(src_dir, dest_dir)


def chown_r(path: str, user: str, group: str) -> None:
    # http://stackoverflow.com/questions/2853723
    for root, dirs, files in os.walk(path):
        for x in dirs:
            shutil.chown(os.path.join(root, x), user, group)
        for x in files:
            shutil.chown(os.path.join(root, x), user, group)


def get_lines_without_comments(filename: str) -> List[str]:
    lines = []
    with open(filename) as f:
        for line in f:
            line = line.partition('#')[0]
            line = line.rstrip()
            line = line.lstrip()
            if line:
                lines.append(line)
    return lines


def moveglob(src: str, dest: str, allow_nothing: bool = False) -> None:
    something = False
    for filename in glob.glob(src):
        shutil.move(filename, dest)
        something = True
    if something or allow_nothing:
        return
    raise ValueError("No files found matching: {}".format(src))


def rmglob(pattern: str) -> None:
    for f in glob.glob(pattern):
        os.remove(f)


def purge(path: str, pattern: str) -> None:
    for f in find(pattern, path):
        log.info("Deleting {}".format(f))
        os.remove(f)


def find(pattern: str, path: str) -> List[str]:
    result = []
    for root, dirs, files in os.walk(path):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                result.append(os.path.join(root, name))
    return result


def find_first(pattern, path):
    try:
        return find(pattern, path)[0]
    except IndexError:
        log.critical('''Couldn't find "{}" in "{}"'''.format(pattern, path))
        raise



#!/usr/bin/env python

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0006_require_contenttypes_0002'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserProfile',
            fields=[
                ('user', models.OneToOneField(related_name='profile', to=settings.AUTH_USER_MODEL, serialize=False, primary_key=True)),  # noqa
                ('per_page', models.PositiveSmallIntegerField(default=50, choices=[(10, '10'), (20, '20'), (50, '50'), (100, '100'), (200, '200'), (500, '500'), (1000, '1000')], verbose_name='Number of items to show per page')),  # noqa
                ('line_length', models.PositiveSmallIntegerField(default=80, verbose_name='Characters to word-wrap text at in results display (0 for no wrap)')),  # noqa
                ('collapse_at', models.PositiveSmallIntegerField(default=400, verbose_name='Number of characters beyond which results field starts collapsed (0 for none)')),  # noqa
                ('is_developer', models.BooleanField(default=False, verbose_name='Enable developer functions?')),  # noqa
                ('title', models.CharField(max_length=20, blank=True)),
                ('address_1', models.CharField(max_length=100, blank=True, verbose_name='Address line 1')),  # noqa
                ('address_2', models.CharField(max_length=100, blank=True, verbose_name='Address line 2')),  # noqa
                ('address_3', models.CharField(max_length=100, blank=True, verbose_name='Address line 3')),  # noqa
                ('address_4', models.CharField(max_length=100, blank=True, verbose_name='Address line 4')),  # noqa
                ('address_5', models.CharField(max_length=100, blank=True, verbose_name='Address line 5 (county)')),  # noqa
                ('address_6', models.CharField(max_length=100, blank=True, verbose_name='Address line 6 (postcode)')),  # noqa
                ('address_7', models.CharField(max_length=100, blank=True, verbose_name='Address line 7 (country)')),  # noqa
                ('telephone', models.CharField(max_length=20, blank=True)),
                ('is_consultant', models.BooleanField(default=False, verbose_name='User is an NHS consultant')),  # noqa
                ('signatory_title', models.CharField(max_length=255, verbose_name='Title for signature (e.g. "Consultant psychiatrist")')),  # noqa
            ],
        ),
    ]



#!/usr/bin/env python
# crate_anon/tools/launch_flower.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import subprocess

from crate_anon.crateweb.config.constants import CELERY_APP_NAME


def main():
    cmdargs = [
        "celery",
        "-A", CELERY_APP_NAME,
        "flower"
    ]
    print("Launching Flower: {}".format(cmdargs))
    subprocess.call(cmdargs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# crate_anon/nlp_manager/nlp_definition.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

# =============================================================================
# Imports
# =============================================================================

import codecs
import datetime
import logging
import os
import sys
from typing import Dict, Iterable, List, Optional

from cardinal_pythonlib.rnc_datetime import get_now_utc_notz
from cardinal_pythonlib.rnc_lang import chunks
from sqlalchemy.engine import Engine
from sqlalchemy.orm.session import Session
from sqlalchemy.schema import MetaData

from crate_anon.anonymise.dbholder import DatabaseHolder
from crate_anon.common.extendedconfigparser import ExtendedConfigParser
from crate_anon.common.sql import TransactionSizeLimiter
from crate_anon.nlp_manager.constants import (
    DEFAULT_MAX_BYTES_BEFORE_COMMIT,
    DEFAULT_MAX_ROWS_BEFORE_COMMIT,
    DEFAULT_TEMPORARY_TABLENAME,
    HashClass,
    MAX_SQL_FIELD_LEN,
    NLP_CONFIG_ENV_VAR,
)

# if sys.version_info.major >= 3 and sys.version_info.minor >= 5:
#     from crate_anon.nlp_manager import input_field_config
#     from crate_anon.nlp_manager import base_nlp_parser  # SEE NEXT LINES

# - see PEP0484 / forward references
# - some circular imports work under Python 3.5 but not 3.4:
#   https://docs.python.org/3/whatsnew/3.5.html#other-language-changes
#   https://bugs.python.org/issue17636
# - see also:
#   http://stackoverflow.com/questions/6351805/cyclic-module-dependencies-and-relative-imports-in-python  # noqa
#   http://stackoverflow.com/questions/35776791/type-hinting-union-with-forward-references  # noqa
# - OK, still problems.
#   Let's strip this back to something sensible.
#   Does BaseNlpParser really need to know about NlpDefinition?
#   - Not directly.
#   - For typing, if it stores a reference (optional).
#   - It could also be given subcomponents instead.
#   Does NlpDefinition really need to know about BaseNlpParser?
#   - Yes, but only for delayed imports.
# - For now, solved by weakening type hints for NlpDefinition.
# - # noinspection PyUnresolvedReferences
#   ... see http://codeoptimism.com/blog/pycharm-suppress-inspections-list/
#   for a full list.

log = logging.getLogger(__name__)


# =============================================================================
# Config class
# =============================================================================

class NlpDefinition(object):
    """
    Class representing NLP master configuration as read from config file.
    """

    # noinspection PyUnresolvedReferences
    def __init__(self, nlpname: str, logtag: str = "") -> None:
        """
        Read config from file.
        """

        # DELAYED IMPORTS (to make life simpler for classes deriving from
        # NlpParser and using NlpDefinition -- they can now do it directly,
        # not just via forward reference).
        from crate_anon.nlp_manager.all_processors import make_processor
        from crate_anon.nlp_manager.input_field_config import InputFieldConfig

        self._nlpname = nlpname
        self._logtag = logtag

        log.info("Loading config for section: {}".format(nlpname))
        # Get filename
        try:
            self._config_filename = os.environ[NLP_CONFIG_ENV_VAR]
            assert self._config_filename
        except (KeyError, AssertionError):
            print(
                "You must set the {} environment variable to point to a CRATE "
                "anonymisation config file. Run crate_print_demo_anon_config "
                "to see a specimen config.".format(NLP_CONFIG_ENV_VAR))
            sys.exit(1)

        # Read config from file.
        self._parser = ExtendedConfigParser()
        self._parser.optionxform = str  # make it case-sensitive
        log.info("Reading config file: {}".format(self._config_filename))
        self._parser.read_file(codecs.open(self._config_filename, "r", "utf8"))

        if not self._parser.has_section(nlpname):
            raise ValueError("No section named {} present".format(nlpname))

        # ---------------------------------------------------------------------
        # Our own stuff
        # ---------------------------------------------------------------------
        self._databases = {}  # type: Dict[str, DatabaseHolder]
        self._progressdb_name = self.opt_str(nlpname, 'progressdb',
                                             required=True)
        self._progdb = self.get_database(self._progressdb_name)
        self._temporary_tablename = self.opt_str(
            nlpname, 'temporary_tablename',
            default=DEFAULT_TEMPORARY_TABLENAME)
        self._hashphrase = self.opt_str(nlpname, 'hashphrase', required=True)
        self._hasher = HashClass(self._hashphrase)
        self._max_rows_before_commit = self.opt_int(
            nlpname, 'max_rows_before_commit', DEFAULT_MAX_ROWS_BEFORE_COMMIT)
        self._max_bytes_before_commit = self.opt_int(
            nlpname, 'max_bytes_before_commit',
            DEFAULT_MAX_BYTES_BEFORE_COMMIT)
        self._now = get_now_utc_notz()

        # ---------------------------------------------------------------------
        # Input field definitions
        # ---------------------------------------------------------------------
        self._inputfielddefs = self.opt_strlist(nlpname, 'inputfielddefs',
                                                required=True, lower=False)
        self._inputfieldmap = {}  # type: Dict[str, InputFieldConfig]
        for x in self._inputfielddefs:
            if x in self._inputfieldmap:
                continue
            self._inputfieldmap[x] = InputFieldConfig(self, x)

        # ---------------------------------------------------------------------
        # NLP processors
        # ---------------------------------------------------------------------
        self._processors = []  # type: List[BaseNlpParser]
        processorpairs = self.opt_strlist(nlpname, 'processors', required=True,
                                          lower=False)
        try:
            for proctype, procname in chunks(processorpairs, 2):
                self.require_section(procname)
                processor = make_processor(proctype, self, procname)
                self._processors.append(processor)
        except:
            log.critical("Bad 'processors' specification")
            raise

        # ---------------------------------------------------------------------
        # Transaction sizes, for early commit
        # ---------------------------------------------------------------------
        self._transaction_limiters = {}  # type: Dict[Session, TransactionSizeLimiter]  # noqa
        # dictionary of session -> TransactionSizeLimiter

    def get_name(self) -> str:
        return self._nlpname

    def get_logtag(self) -> str:
        return self._logtag

    def get_parser(self) -> ExtendedConfigParser:
        return self._parser

    def hash(self, text: str) -> str:
        return self._hasher.hash(text)

    def get_temporary_tablename(self) -> str:
        return self._temporary_tablename

    def set_echo(self, echo: bool) -> None:
        self._progdb.engine.echo = echo
        for db in self._databases.values():
            db.engine.echo = echo
        # Now, SQLAlchemy will mess things up by adding an additional handler.
        # So, bye-bye:
        for logname in ('sqlalchemy.engine.base.Engine',
                        'sqlalchemy.engine.base.OptionEngine'):
            logger = logging.getLogger(logname)
            logger.handlers = []  # type: List[logging.Handler]

    def require_section(self, section: str) -> None:
        if not self._parser.has_section(section):
            msg = "Missing config section: {}".format(section)
            log.critical(msg)
            raise ValueError(msg)

    def opt_str(self, section: str, option: str, required: bool = False,
                default: str = None) -> str:
        return self._parser.get_str(section, option, default=default,
                                    required=required)

    def opt_strlist(self, section: str, option: str, required: bool = False,
                    lower: bool = True, as_words: bool = True) -> List[str]:
        return self._parser.get_str_list(section, option, as_words=as_words,
                                         lower=lower, required=required)

    def opt_int(self, section: str, option: str,
                default: Optional[int]) -> Optional[int]:
        return self._parser.getint(section, option, fallback=default)

    def opt_bool(self, section: str, option: str, default: bool) -> bool:
        return self._parser.getboolean(section, option, fallback=default)

    def get_database(self, name_and_cfg_section: str,
                     with_session: bool = True,
                     with_conn: bool = False,
                     reflect: bool = False) -> DatabaseHolder:
        if name_and_cfg_section in self._databases:
            return self._databases[name_and_cfg_section]
        assert len(name_and_cfg_section) <= MAX_SQL_FIELD_LEN
        db = self._parser.get_database(name_and_cfg_section,
                                       with_session=with_session,
                                       with_conn=with_conn,
                                       reflect=reflect)
        self._databases[name_and_cfg_section] = db
        return db

    def get_env_dict(self, section: str,
                     parent_env: Optional[Dict]=None) -> Dict:
        return self._parser.get_env_dict(section, parent_env=parent_env)

    def get_progdb_session(self) -> Session:
        return self._progdb.session

    def get_progdb_engine(self) -> Engine:
        return self._progdb.engine

    def get_progdb_metadata(self) -> MetaData:
        return self._progdb.metadata

    def commit_all(self) -> None:
        """
        Execute a COMMIT on all databases (destination + progress).
        """
        self.commit(self.get_progdb_session())
        for db in self._databases.values():
            self.commit(db.session)

    def get_transation_limiter(self,
                               session: Session) -> TransactionSizeLimiter:
        if session not in self._transaction_limiters:
            self._transaction_limiters[session] = TransactionSizeLimiter(
                session,
                max_rows_before_commit=self._max_rows_before_commit,
                max_bytes_before_commit=self._max_bytes_before_commit)
        return self._transaction_limiters[session]

    def notify_transaction(self, session: Session,
                           n_rows: int, n_bytes: int,
                           force_commit: bool=False) -> None:
        tl = self.get_transation_limiter(session)
        tl.notify(n_rows=n_rows, n_bytes=n_bytes, force_commit=force_commit)

    def commit(self, session: Session) -> None:
        tl = self.get_transation_limiter(session)
        tl.commit()

    # noinspection PyUnresolvedReferences
    def get_processors(self) -> List['base_nlp_parser.BaseNlpParser']:  # typing / circular reference problem  # noqa
        return self._processors

    # noinspection PyUnresolvedReferences
    def get_ifconfigs(self) -> Iterable['input_field_config.InputFieldConfig']:  # typing / circular reference problem  # noqa
        return self._inputfieldmap.values()

    def get_now(self) -> datetime.datetime:
        return self._now



#!/usr/bin/env python
# crate_anon/anonymise/anon_constants.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Shared constants for CRATE anonymiser.
"""

from enum import unique

from sqlalchemy import BigInteger, Integer
from cardinal_pythonlib.rnc_lang import StrEnum

from crate_anon.version import VERSION, VERSION_DATE

# =============================================================================
# Logging
# =============================================================================

LOG_FORMAT = '%(asctime)s.%(msecs)03d:%(levelname)s:%(name)s:%(message)s'
LOG_DATEFMT = '%Y-%m-%d %H:%M:%S'

LOG_COLORS = {
    'DEBUG': 'cyan',
    'INFO': 'green',
    'WARNING': 'yellow',
    'ERROR': 'red',
    'CRITICAL': 'red,bg_white',
}

# =============================================================================
# Cosmetic
# =============================================================================

BIGSEP = "=" * 20 + " "
SEP = "-" * 20 + " "

# =============================================================================
# Defaults for command-line options
# =============================================================================

DEFAULT_REPORT_EVERY = 100000  # 100k
DEFAULT_CHUNKSIZE = 100000  # 100k

# =============================================================================
# Environment
# =============================================================================

CONFIG_ENV_VAR = 'CRATE_ANON_CONFIG'

# =============================================================================
# Data dictionary
# =============================================================================

DATEFORMAT_ISO8601 = "%Y-%m-%dT%H:%M:%S%z"  # e.g. 2013-07-24T20:04:07+0100
DEFAULT_INDEX_LEN = 20  # for data types where it's mandatory
DEFAULT_MAX_ROWS_BEFORE_COMMIT = 1000
DEFAULT_MAX_BYTES_BEFORE_COMMIT = 80 * 1024 * 1024

LONGTEXT = "LONGTEXT"

MAX_PID_STR = "9" * 10  # e.g. NHS numbers are 10-digit

# Better overall than string.maketrans:
ODD_CHARS_TRANSLATE = [chr(x) for x in range(0, 256)]
for c in '()/ ':
    ODD_CHARS_TRANSLATE[ord(c)] = '_'
for i in range(0, 32):
    ODD_CHARS_TRANSLATE[i] = '_'
for i in range(127, 256):
    ODD_CHARS_TRANSLATE[i] = '_'
ODD_CHARS_TRANSLATE = "".join(ODD_CHARS_TRANSLATE)

PidType = BigInteger
TridType = Integer
MAX_TRID = 2 ** 31 - 1
# https://dev.mysql.com/doc/refman/5.0/en/numeric-type-overview.html
# Maximum INT UNSIGNED is              4294967295 == 2 ** 32 - 1.
# INT range is                        -2147483648 == -(2 **  31) to
#                                     +2147483647 == 2 ** 31 - 1
# ... note that this is inadequate for 10-digit NHS numbers.
# Maximum BIGINT UNSIGNED is 18446744073709551615 == 2 ** 64 - 1.
# BIGINT range is            -9223372036854775808 == -(2 ** 63) to
#                            +9223372036854775807 == 2 ** 64 - 1


@unique
class ALTERMETHOD(StrEnum):
    TRUNCATEDATE = "truncate_date"
    SCRUBIN = "scrub"
    BINARY_TO_TEXT = "binary_to_text"
    FILENAME_TO_TEXT = "filename_to_text"
    FILENAME_FORMAT_TO_TEXT = "filename_format_to_text"  # new in v0.18.18
    SKIP_IF_TEXT_EXTRACT_FAILS = "skip_if_extract_fails"
    # HTML_ESCAPE = "html_escape"
    HTML_UNESCAPE = "html_unescape"
    HTML_UNTAG = "html_untag"


@unique
class DECISION(StrEnum):
    OMIT = "OMIT"
    INCLUDE = "include"


@unique
class INDEX(StrEnum):
    NORMAL = "I"
    UNIQUE = "U"
    FULLTEXT = "F"


@unique
class SCRUBMETHOD(StrEnum):
    WORDS = "words"
    PHRASE = "phrase"
    NUMERIC = "number"
    DATE = "date"
    CODE = "code"


@unique
class SCRUBSRC(StrEnum):
    PATIENT = "patient"
    THIRDPARTY = "thirdparty"
    THIRDPARTY_XREF_PID = "thirdparty_xref_pid"


@unique
class SRCFLAG(StrEnum):
    PK = "K"
    ADD_SRC_HASH = "H"
    PRIMARY_PID = "P"
    DEFINES_PRIMARY_PIDS = "*"
    MASTER_PID = "M"
    CONSTANT = "C"
    ADDITION_ONLY = "A"
    OPT_OUT = "!"
    REQUIRED_SCRUBBER = "R"


# =============================================================================
# Databases
# =============================================================================

CHARSET = 'utf8'
TABLE_KWARGS = {
    # MySQL:
    'mysql_charset': CHARSET,
    'mysql_engine': 'InnoDB',
}

MAX_IDENTIFIER_LENGTH = 64
# MySQL: 64 -- http://dev.mysql.com/doc/refman/5.7/en/identifiers.html

# =============================================================================
# Demo config
# =============================================================================

# noinspection PyPep8
DEMO_CONFIG = r"""# Configuration file for CRATE anonymiser (crate_anonymise).
# Version {VERSION} ({VERSION_DATE}).
#
# Boolean values can be 0/1, Y/N, T/F, True/False.

# =============================================================================
# Main settings
# =============================================================================

[main]

# -----------------------------------------------------------------------------
# Data dictionary
# -----------------------------------------------------------------------------
# Specify a data dictionary in TSV (tab-separated value) format, with a header
# row.
# Columns in the data dictionary (which can be in any order as long as the
# header row matches the data):
#
# src_db
#     Specify the source database.
#     Database names are those used in source_databases list below; they
#     don't have to be SQL database names.
# src_table
#     Table name in source database.
# src_field
#     Field name in source database.
# src_datatype
#     SQL data type in source database, e.g. INT, VARCHAR(50).
#
# src_flags
#     One or more of the following characters:
#
#     {SRCFLAG.PK}
#         PK
#         This field is the primary key (PK) for the table it's in.
#
#     {SRCFLAG.ADD_SRC_HASH}
#         ADD SOURCE HASH
#         Add source hash of the record, for incremental updates?
#         - This flag may only be set for src_pk fields (which cannot then be
#           omitted in the destination, and which require the index={INDEX.UNIQUE}
#           setting, so that a unique index is created for this field).
#         - If set, a field is added to the destination table, with field
#           name as set by the config's source_hash_fieldname variable,
#           containing a hash of the contents of the source record -- all
#           fields that are not omitted, OR contain scrubbing information
#           (scrub_src). The field is of type VARCHAR and its length is
#           determined by the hash_method parameter (see below).
#         - This table is then capable of incremental updates.
#
#     {SRCFLAG.CONSTANT}
#         CONSTANT
#         Record contents are constant (will not change) for a given PK.
#         - An alternative to '{SRCFLAG.ADD_SRC_HASH}'. Can't be used with it.
#         - The flag can be set only on src_pk fields, which can't be omitted
#           in the destination, and which have the same index requirements as
#           the '{SRCFLAG.ADD_SRC_HASH}' flag.
#         - If set, no hash is added to the destination, but the destination
#           contents are assumed to exist and not to have changed.
#         - Be CAUTIOUS with this flag, i.e. certain that the contents will
#           not change.
#         - Intended for very data-intensive fields, such as BLOB fields
#           containing binary documents, where hashing would be quite slow
#           over many gigabytes of data.
#         - Does not imply that the whole table cannot change!
#
#     {SRCFLAG.ADDITION_ONLY}
#         ADDITION ONLY
#         Addition only. It is assumed that records can only be added, not
#         deleted.
#
#     {SRCFLAG.PRIMARY_PID}
#         PRIMARY PID
#         Primary patient ID field. If set,
#         (a) This field will be used to link records for the same patient
#             across all tables. It must therefore be present, and marked in
#             the data dictionary, for ALL tables that contain patient-
#             identifiable information.
#         (b) If the field is not omitted: the field will be hashed as the
#             primary ID (database patient primary key) in the destination,
#             and a transient research ID (TRID) also added.
#
#     {SRCFLAG.DEFINES_PRIMARY_PIDS}
#         DEFINES PRIMARY PIDS.
#         This field *defines* primary PIDs. If set, this row will be used to
#         search for all patient IDs, and will define them for this database.
#         Only those patients will be processed (for all tables containing
#         patient info). Typically, this flag is applied to a SINGLE field in a
#         SINGLE table, usually the principal patient registration/demographics
#         table.
#
#     {SRCFLAG.MASTER_PID}
#         MASTER PID
#         Master ID (e.g. NHS number).
#         The field will be hashed with the master PID hasher.
#
#     {SRCFLAG.OPT_OUT}
#         OPT OUT
#         This field is used to mark that the patient wishes to opt out
#         entirely. It must be in a table that also has a primary patient ID
#         field (because that's the ID that will be omitted). If the opt-out
#         field contains a value that's defined in the optout_col_values
#         setting (see below), that patient will be opted out entirely from
#         the anonymised database.
#
#     {SRCFLAG.REQUIRED_SCRUBBER}
#         REQUIRED SCRUBBER
#         If this field is a scrub_src field (see below), and this flag is set,
#         then at least one non-NULL value for this field must be present for
#         each patient, or no information will be processed for this patient.
#         (Typical use: where you have a master patient index separate from the
#         patient name table, and data might have been brought across
#         partially, so there are some missing names. In this situation, text
#         might go unscrubbed because the names are missing. Setting this flag
#         for the name field will prevent this.)
#
# scrub_src
#     One of:
#           "{SCRUBSRC.PATIENT}",
#           "{SCRUBSRC.THIRDPARTY}",
#           "{SCRUBSRC.THIRDPARTY_XREF_PID}",
#           or blank.
#     Explanations:
#     - "{SCRUBSRC.PATIENT}":
#       Contains patient-identifiable information that must be removed from
#       "scrub_in" fields.
#     - "{SCRUBSRC.THIRDPARTY}":
#       Contains identifiable information about carer/family/other third party,
#       which must be removed from "scrub_in" fields.
#     - "{SCRUBSRC.THIRDPARTY_XREF_PID}":
#       This field is a patient identifier for ANOTHER patient (such as a
#       relative). The scrubber should recursively include THAT patient's
#       identifying information as third-party information for THIS patient.
#
# scrub_method
#     Applicable to scrub_src fields. Manner in which this field should be
#     treated for scrubbing.
#     Options:
#
#     - "{SCRUBMETHOD.WORDS}"
#       Treat as a set of textual words. This is the default for all textual
#       fields (e.g. CHAR, VARCHAR, TEXT). Typically used for names.
#       Also OK for e-mail addresses.
#
#     - "{SCRUBMETHOD.PHRASE}"
#       Treat as a textual phrase (a sequence of words to be replaced only when
#       they occur in sequence). Typically used for address components.
#
#     - "{SCRUBMETHOD.NUMERIC}"
#       Treat as a number. This is the default for all numeric fields (e.g.
#       INTEGER, FLOAT). If you have a phone number in a text field, use this
#       method; it will be scrubbed regardless of spacing/punctuation.
#
#     - "{SCRUBMETHOD.CODE}"
#       Teat as an alphanumeric code. Suited to postcodes. Very like the
#       numeric method, but permits non-digits.
#
#     - "{SCRUBMETHOD.DATE}"
#       Treat as a date. This is the default for all DATE/DATETIME fields.
#
# decision
#     One of:
#     - "{DECISION.OMIT}": omit the field from the output entirely;
#     - "{DECISION.INCLUDE}": include it.
#     This is case sensitive, for safety.
#
# inclusion_values
#     - Either blank, or an expression that evaluates to a Python iterable
#       (e.g. list or tuple) with Python's ast.literal_eval() function (see
#       https://docs.python.org/3.4/library/ast.html).
#     - Examples:
#           [None, 0]
#           [True, 1, 'yes', 'true', 'Yes', 'True']
#     - If this is not blank/None, then it serves as a ROW INCLUSION LIST -
#       the source row will only be processed if the field's value is one of
#       the inclusion values.
#     - It applies to the raw value from the database (before any
#       transformation via alter_method).
#     - This is not applied to scrub_src fields (which contribute to the
#       scrubber regardless.
#     - Note that "[None]" is a list with one member, None, whereas "None"
#       is equivalent to leaving the field blank.
#
# exclusion_values
#     - As for inclusion_values, but the row is excluded if the field's value
#       is in the exclusion_values list.
#
# alter_method
#     Manner in which to alter the data. Blank, or one or more of:
#
#     - "{ALTERMETHOD.SCRUBIN}"
#       Scrub in. Applies to text fields only. The field will have its contents
#       anonymised (using information from other fields). Use this for any
#       text field that end users might store free-text comments in.
#
#     - "{ALTERMETHOD.TRUNCATEDATE}"
#       Truncate this date to the first of the month. Applicable to text or
#       date-as-text fields.
#
#     - "{ALTERMETHOD.BINARY_TO_TEXT}=EXTFIELDNAME"
#       Convert a binary field (e.g. VARBINARY, BLOB) to text (e.g. LONGTEXT).
#       The binary data is taken to be the representation of a document.
#       The field EXTFIELDNAME, which must be in the same source table, must
#       contain the file extension (e.g. "pdf", ".pdf") or a filename with that
#       extension (e.g. "/some/path/mything.pdf"), so that the anonymiser knows
#       how to treat the binary data to extract text from it.
#
#     - "{ALTERMETHOD.FILENAME_TO_TEXT}"
#       As for the binary-to-text option, but the field contains a filename
#       (the contents of which is converted to text), rather than containing
#       binary data directly.
#
#     - "{ALTERMETHOD.FILENAME_FORMAT_TO_TEXT}=SOMEFORMAT"
#       A more powerful way of specifying a filename that can be created using
#       data from this table. The SOMEFORMAT parameter is an unquoted Python 
#       str.format() string; see 
#       https://docs.python.org/3.4/library/stdtypes.html#str.format .
#       The dictionary passed in to format() is created from all fields in the 
#       row.
#
#       Using an example from RiO: if your ClientDocuments table contains a 
#       ClientID column (e.g. '999999') and a Path column (e.g. 
#       'appointment_letter.pdf'), and you know that the actual file will then
#       be found at 'C:\some\path\999999\docs\appointment_letter.pdf', then you
#       can specify this with
#           {ALTERMETHOD.FILENAME_FORMAT_TO_TEXT}=C:\some\path\{{ClientID}}\docs\{{Path}}
#       You probably want to apply this alter_method to the Path column in this
#       example, though that's not mandatory.
#
#     - "{ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS}"
#       If one of the text extraction methods is specified, and this flag is
#       also specified, then the data row will be skipped if text extrcation
#       fails (rather than inserted with a NULL value for the text). This is
#       helpful, for example, if your text-processing pipeline breaks; the
#       option prevents rows being created erroneously with NULL text values,
#       so that a subsequent incremental update will fix the problems once
#       you've fixed your text extraction tools.
#
#     - "{ALTERMETHOD.HTML_UNESCAPE}"
#       HTML encoding is removed, e.g. convert "&amp;" to "&" and "&lt;" to "<"
#
#     - "{ALTERMETHOD.HTML_UNTAG}"
#       HTML tags are removed, e.g. from
#           <a href="http://somewhere">see link</a>
#       to
#           see link
#
#     You can specify multiple options separated by commas.
#     Not all are compatible (e.g. scrubbing is for text; date truncation is
#     for dates).
#     If there's more than one, text extraction from BLOBs/files is performed
#     first. After that, they are executed in sequence. (The position of the
#     skip-if-text-extraction-fails flag is immaterial.)
#     A typical combination might be:
#           {ALTERMETHOD.FILENAME_TO_TEXT},{ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS},{ALTERMETHOD.SCRUBIN}
#     or:
#           {ALTERMETHOD.HTML_UNTAG},{ALTERMETHOD.HTML_UNESCAPE},{ALTERMETHOD.SCRUBIN}
#
# dest_table
#     Table name in destination database.
# dest_field
#     Field name in destination database.
# dest_datatype
#     SQL data type in destination database.
#     If omitted, the source SQL data type is translated appropriately.
# index
#     One of:
#     - blank: no index.
#     - "{INDEX.NORMAL}"
#       ... create a normal index on the destination field.
#     - "{INDEX.UNIQUE}"
#       ... create a unique index on the destination field.
#     - "{INDEX.FULLTEXT}"
#       ... create a FULLTEXT index, for rapid searching within long text
#       fields. Only applicable to one field per table.
# indexlen
#     Integer. Can be blank. If not, sets the prefix length of the index.
#     Mandatory in MySQL if you apply a normal (+/- unique) index to a TEXT
#     or BLOB field. Not required for FULLTEXT indexes.
# comment
#     Field comment, stored in destination database.

data_dictionary_filename = testdd.tsv

# -----------------------------------------------------------------------------
# Encryption phrases/passwords
# -----------------------------------------------------------------------------

    # PID-to-RID hashing method. Options are:
    # - HMAC_MD5 - produces a 32-character digest
    # - HMAC_SHA256 - produces a 64-character digest
    # - HMAC_SHA512 - produces a 128-character digest
hash_method = HMAC_MD5

per_table_patient_id_encryption_phrase = SOME_PASSPHRASE_REPLACE_ME

master_patient_id_encryption_phrase = SOME_OTHER_PASSPHRASE_REPLACE_ME

change_detection_encryption_phrase = YETANOTHER

# -----------------------------------------------------------------------------
# Text extraction
# -----------------------------------------------------------------------------

    # extract_text_extensions_permitted and extract_text_extensions_prohibited
    # govern what kinds of files are accepted for text extraction. It is very
    # likely that you'll want to apply such restrictions; for example, if your
    # database contains .jpg files, it's a waste of trying to extract text from
    # them (and in theory, if your text extraction tool provided sufficient
    # detail, such as binary-encoding the JPEG, you might leak identifiable
    # information, such as a photo).
    #
    # - The "permitted" and "prohibited" settings are both lists of strings.
    # - If the "permitted" list is not empty then a file will be processed
    #   only if its extension is in the permitted list. Otherwise, it will be
    #   processed only if it is not in the prohibited list.
    # - The extensions must include the "." prefix.
    # - Case sensitivity is controlled by the extra flag.

extract_text_extensions_case_sensitive = False
extract_text_extensions_permitted =
extract_text_extensions_prohibited =

    # Use the plainest possible layout for text extraction?
    # False = better for human layout. Table example from DOCX:
    #     +---------+---------+
    #     | AAA AAA | BBB BBB |
    #     | AAA AAA | BBB BBB |
    #     +---------+---------+
    # True = good for natural language processing. Table example from DOCX:
    #     ---------------------
    #       AAA AAA
    #       AAA AAA
    #     ---------------------
    #                 BBB BBB
    #                 BBB BBB
    #     ---------------------
    # ... note the absence of vertical interruptions, and that text from one
    # cell remains contiguous.
extract_text_plain = False

    # Default width to word-wrap extracted text to
extract_text_width = 80

# -----------------------------------------------------------------------------
# Anonymisation
# -----------------------------------------------------------------------------

    # Patient information will be replaced with this. For example, XXXXXX or
    # [___] or [__PPP__] or [__ZZZ__]; the bracketed forms can be a bit easier
    # to spot, and work better if they directly abut other text.
replace_patient_info_with = [__PPP__]

    # Third-party information will be replaced by this.
    # For example, YYYYYY or [...] or [__TTT__] or [__QQQ__].
replace_third_party_info_with = [__TTT__]

    # For fields marked as scrub_src = {SCRUBSRC.THIRDPARTY_XREF_PID},
    # how deep should we recurse? The default is 1. Beware making this too
    # large; the recursion trawls a lot of information (and also uses an
    # extra simultaneous database cursor for each recursion).
thirdparty_xref_max_depth = 1

    # Things to be removed irrespective of patient-specific information will be
    # replaced by this (for example, if you opt to remove all things looking
    # like telephone numbers). For example, ZZZZZZ or [~~~].
replace_nonspecific_info_with = [~~~]

    # Strings to append to every "scrub from" string.
    # For example, include "s" if you want to scrub "Roberts" whenever you
    # scrub "Robert".
    # Applies to {SCRUBMETHOD.WORDS}, but not to {SCRUBMETHOD.PHRASE}.
    # Multiline field: https://docs.python.org/2/library/configparser.html
scrub_string_suffixes =
    s

    # Specify maximum number of errors (insertions, deletions, substitutions)
    # in string regex matching. Beware using a high number! Suggest 1-2.
string_max_regex_errors = 1

    # Is there a minimum length to apply string_max_regex_errors? For example,
    # if you allow one typo and someone is called Ian, all instances of 'in' or
    # 'an' will be wiped. Note that this apply to scrub-source data.
min_string_length_for_errors = 4

    # Is there a minimum length of string to scrub WITH? For example, if you
    # specify 2, you allow two-letter names such as Al to be scrubbed, but you
    # allow initials through, and therefore prevent e.g. 'A' from being
    # scrubbed from the destination. Note that this applies to scrub-source
    # data.
min_string_length_to_scrub_with = 2

    # WHITELIST.
    # Are there any words not to scrub? For example, "the", "road", "street"
    # often appear in addresses, but you might not want them removed. Be
    # careful in case these could be names (e.g. "Lane").
    # Specify these as a list of FILENAMES, where the files contain words; e.g.
    #
    # whitelist_filenames = /some/path/short_english_words.txt
    #
    # Here's a suggestion for some of the sorts of words you might include:
    #     am
    #     an
    #     as
    #     at
    #     bd
    #     by
    #     he
    #     if
    #     is
    #     it
    #     me
    #     mg
    #     od
    #     of
    #     on
    #     or
    #     re
    #     so
    #     to
    #     us
    #     we
    #     her
    #     him
    #     tds
    #     she
    #     the
    #     you
    #     road
    #     street
whitelist_filenames =

    # BLACKLIST
    # Are there any words you always want to remove?
    # Specify these as a list of filenames, e.g
    #
    # blacklist_filenames = /some/path/boy_names.txt
    #     /some/path/girl_names.txt
    #     /some/path/common_surnames.txt
blacklist_filenames =

    # Nonspecific scrubbing of numbers of a certain length?
    # For example, scrubbing all 11-digit numbers will remove modern UK
    # telephone numbers in conventional format. To do this, specify
    # scrub_all_numbers_of_n_digits = 11. You could scrub both 10- and 11-digit
    # numbers by specifying both numbers (in multiline format, as above);
    # 10-digit numbers would include all NHS numbers. Avoid using this for
    # short numbers; you may lose valuable numeric data!
scrub_all_numbers_of_n_digits =

    # Nonspecific scrubbing of UK postcodes?
    # See https://www.mrs.org.uk/pdf/postcodeformat.pdf ; these can look like
    # FORMAT    EXAMPLE
    # AN NAA    M1 1AA
    # ANN NAA   M60 1NW
    # AAN NAA   CR2 6XH
    # AANN NAA  DN55 1PT
    # ANA NAA   W1A 1HQ
    # AANA NAA  EC1A 1BB
scrub_all_uk_postcodes = False

    # Anonymise at word boundaries? True is more conservative; False is more
    # liberal and will deal with accidental word concatenation. With ID
    # numbers, beware if you use a prefix, e.g. if people write 'M123456' or
    # 'R123456'; in that case you will need
    #       anonymise_numbers_at_word_boundaries_only = False.
anonymise_codes_at_word_boundaries_only = True
    # ... applies to {SCRUBMETHOD.CODE}
anonymise_dates_at_word_boundaries_only = True
    # ... applies to {SCRUBMETHOD.DATE}
anonymise_numbers_at_word_boundaries_only = False
    # ... applies to {SCRUBMETHOD.NUMERIC}
anonymise_numbers_at_numeric_boundaries_only = True
    # ... applies to {SCRUBMETHOD.NUMERIC}
    # ... if True, will not scrub "234" from "123456"
    # ... setting this to False is extremely conservative
anonymise_strings_at_word_boundaries_only = True
    # ... applies to {SCRUBMETHOD.WORDS} and {SCRUBMETHOD.PHRASE}

# -----------------------------------------------------------------------------
# Output fields and formatting
# -----------------------------------------------------------------------------

    # Name used for the primary patient ID in the mapping table.
mapping_patient_id_fieldname = patient_id

    # Research ID field name. This will be a VARCHAR of length determined by
    # hash_method. Used to replace per_table_patient_id_field.
research_id_fieldname = brcid

    # Transient integer research ID (TRID) fieldname.
    # An unsigned integer field with this name will be added to every table
    # containing a primary patient ID (in the source) or research ID (in the
    # destination).
trid_fieldname = trid

    # Name used for the master patient ID in the mapping table.
mapping_master_id_fieldname = nhsnum

    # Similarly, used to replace ddgen_master_pid_fieldname:
master_research_id_fieldname = nhshash

    # Change-detection hash fieldname. This will be a VARCHAR of length
    # determined by hash_method.
source_hash_fieldname = _src_hash

    # Date-to-text conversion formats
date_to_text_format = %Y-%m-%d
    # ... ISO-8601, e.g. 2013-07-24
datetime_to_text_format = %Y-%m-%dT%H:%M:%S
    # ... ISO-8601, e.g. 2013-07-24T20:04:07

    # Append source table/field to the comment? Boolean.
append_source_info_to_comment = True

# -----------------------------------------------------------------------------
# Destination database configuration
# See the [destination_database] section for connection details.
# -----------------------------------------------------------------------------

    # Specify the maximum number of rows to be processed before a COMMIT is
    # issued on the database transaction. This prevents the transaction growing
    # too large.
    # Default is {DEFAULT_MAX_ROWS_BEFORE_COMMIT}.
max_rows_before_commit = {DEFAULT_MAX_ROWS_BEFORE_COMMIT}

    # Specify the maximum number of source-record bytes (approximately!) that
    # are processed before a COMMIT is issued on the database transaction. This
    # prevents the transaction growing too large. The COMMIT will be issued
    # *after* this limit has been met/exceeded, so it may be exceeded if the
    # transaction just before the limit takes the cumulative total over the
    # limit.
    # Default is {DEFAULT_MAX_BYTES_BEFORE_COMMIT}.
max_bytes_before_commit = {DEFAULT_MAX_BYTES_BEFORE_COMMIT}

    # We need a temporary table name for incremental updates. This can't be the
    # name of a real destination table. It lives in the destination database.
temporary_tablename = _temp_table

# -----------------------------------------------------------------------------
# Choose databases (defined in their own sections).
# -----------------------------------------------------------------------------

    # Source database list. Can be lots.
source_databases =
    mysourcedb1
    mysourcedb2

    # Destination database. Just one.
destination_database = my_destination_database

    # Admin database. Just one.
admin_database = my_admin_database

# -----------------------------------------------------------------------------
# PROCESSING OPTIONS, TO LIMIT DATA QUANTITY FOR TESTING
# -----------------------------------------------------------------------------

    # Limit the number of patients to be processed? Specify 0 (the default) for
    # no limit.
debug_max_n_patients =

    # Specify a list of integer patient IDs, for debugging? If specified, this
    # list will be used directly (overriding the patient ID source specified in
    # the data dictionary, and overriding debug_max_n_patients).
debug_pid_list =

# -----------------------------------------------------------------------------
# Opting out entirely
# -----------------------------------------------------------------------------

# Patients who elect to opt out entirely have their PIDs stored in the OptOut
# table of the admin database. ENTRIES ARE NEVER REMOVED FROM THIS LIST BY
# CRATE. It can be populated in three ways:
#   1. Manually, by adding a PID to the column opt_out.pid).
#   2. By maintaining a text file list of integer PIDs. Any PIDs in this file
#      are added to the opt-out list.
#   3. By flagging a source database field as indicating an opt-out, using the
#      src_flags = "{SRCFLAG.OPT_OUT}" marker.

    # If you set this, each line of the file(s) is scanned for an integer,
    # taken to the PID of a patient who wishes to opt out.
optout_pid_filenames =

    # If you set this, each line of the file(s) is scanned for an integer,
    # taken to the MPID of a patient who wishes to opt out.
optout_mpid_filenames =

    # If you mark a field in the data dictionary as an opt-out field (see
    # above), that says "the field tells you whether the patient opts out or
    # not". But is it "opt out" or "not"? If the actual value matches one
    # below, then it's "opt out". Specify a LIST OF PYTHON VALUES; for example:
    #       optout_col_values = [True, 1, '1', 'Yes', 'yes', 'Y', 'y']
optout_col_values =

# =============================================================================
# Destination database details. User should have WRITE access.
# =============================================================================
# Use SQLAlchemy URLs: see
#       http://docs.sqlalchemy.org/en/latest/core/engines.html
# You may need to install additional drivers, e.g.
#       pip install SOME_DRIVER
# ... see the documentation.

[my_destination_database]

url = mysql+mysqldb://username:password@127.0.0.1:3306/output_databasename?charset=utf8

# =============================================================================
# Administrative database. User should have WRITE access.
# =============================================================================

[my_admin_database]

url = mysql+mysqldb://username:password@127.0.0.1:3306/admin_databasename?charset=utf8

# =============================================================================
# SOURCE DATABASE DETAILS BELOW HERE.
# User should have READ access only for safety.
# =============================================================================

# -----------------------------------------------------------------------------
# Source database example 1
# -----------------------------------------------------------------------------

[mysourcedb1]

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # CONNECTION DETAILS
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

url = mysql+mysqldb://username:password@127.0.0.1:3306/source_databasename?charset=utf8

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # INPUT FIELDS, FOR THE AUTOGENERATION OF DATA DICTIONARIES
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # - For field specifications, fields can either be specified as "column"
    #   (to match any table) or "table.column", to match a specific table.
    #   They are case-insensitive.
    #   Wildcards (*, ?) may also be used (as per Python's fnmatch).

    # By default, most fields (except PKs and patient ID codes) are marked
    # as "OMIT", pending human review. If you want to live dangerously, set
    # this to False, and they will be marked as "include" from the outset.
ddgen_omit_by_default = True

    # You can specify additional fields to omit...
ddgen_omit_fields =

    # ... or include. "Omit" overrides "include".
    # If a field contains scrubbing source information, it will also be omitted
    # pending human review, regardless of other settings.
ddgen_include_fields =

    # Allow the absence of patient info? Used to copy databases; WILL NOT
    # ANONYMISE. Boolean; default is False.
ddgen_allow_no_patient_info = False

    # Specify the (typically integer) patient identifier present in EVERY
    # table. It will be replaced by the research ID in the destination
    # database.
ddgen_per_table_pid_field = patient_id

    # Add every instance of a per-table PID field to the patient scrubber?
    # This is a very conservative setting, and should be unnecessary as the
    # single master "PID-defining" column (see ddgen_pid_defining_fieldnames)
    # should be enough.
    # (Note that per-table PIDs are always replaced by RIDs - this setting
    # governs whether the scrubber used to scrub free-text fields also
    # works through every single per-table PID).
ddgen_add_per_table_pids_to_scrubber = False

    # Master patient ID fieldname. Used for e.g. NHS numbers.
ddgen_master_pid_fieldname = nhsnum

    # Blacklist any tables when creating new data dictionaries?
    # This is case-insensitive, and you can use */? wildcards (as per Python's
    # fnmatch module).
ddgen_table_blacklist =

    # Whitelist any tables? (Whitelists override blacklists.)
ddgen_table_whitelist =

    # List any fields that all tables MUST contain. If a table doesn't contain
    # all of the field(s) listed here, it will be skipped.
ddgen_table_require_field_absolute =

    # List any fields that are required conditional on other fields.
    # List them as one or more pairs: "A, B" where B is required if A is
    # present (or the table will be skipped).
ddgen_table_require_field_conditional =

    # Blacklist any fields (regardless of their table) when creating new data
    # dictionaries? Wildcards of */? operate as above.
ddgen_field_blacklist =

    # Whitelist any fields? (Whitelists override blacklists.)
ddgen_field_whitelist =

    # Fieldnames assumed to be their table's PK:
ddgen_pk_fields =

    # Assume that content stays constant?
    # (Applies {SRCFLAG.CONSTANT} to PK fields; q.v.)
    # This is the default; then ddgen_constant_content_tables and
    # ddgen_nonconstant_content_tables can override (of which,
    # ddgen_nonconstant_content_tables takes priority if a table matches both).
ddgen_constant_content = False

    # Table-specific overrides for ddgen_constant_content, as above.
ddgen_constant_content_tables =
ddgen_nonconstant_content_tables =

    # Assume that records can only be added, not deleted?
ddgen_addition_only = False

    # Table-specific overrides for ddgen_addition_only, similarly.
ddgen_addition_only_tables =
ddgen_deletion_possible_tables =

    # Predefine field(s) that define the existence of patient IDs? UNUSUAL.
ddgen_pid_defining_fieldnames =

    # Default fields to scrub from
ddgen_scrubsrc_patient_fields =
ddgen_scrubsrc_thirdparty_fields =
ddgen_scrubsrc_thirdparty_xref_pid_fields =

    # Are any scrub_src fields required (mandatory), i.e. must have non-NULL
    # data in at least one row?
ddgen_required_scrubsrc_fields =

    # Override default scrubbing methods
ddgen_scrubmethod_code_fields =
ddgen_scrubmethod_date_fields =
ddgen_scrubmethod_number_fields =
ddgen_scrubmethod_phrase_fields =

    # Known safe fields, exempt from scrubbing
ddgen_safe_fields_exempt_from_scrubbing =

    # Define minimum text field length for scrubbing (shorter is assumed safe)
ddgen_min_length_for_scrubbing = 4

    # Other default manipulations
ddgen_truncate_date_fields =

    # Fields containing filenames, which files should be converted to text
ddgen_filename_to_text_fields =

    # Fields containing raw binary data from files (binary large objects;
    # BLOBs), whose contents should be converted to text -- paired with fields
    # in the same table containing their file extension (e.g. "pdf", ".PDF") or
    # a filename having that extension.
    # Specify it as a list of comma-joined pairs, e.g.
    #     ddgen_binary_to_text_field_pairs = binary1field, ext1field
    #         binary2field, ext2field
    #         ...
    # The first (binaryfield) can be specified as column or table.column,
    # but the second must be column only.
ddgen_binary_to_text_field_pairs =

    # Specify any text-extraction rows for which you also want to set the flag
    # "{ALTERMETHOD.SKIP_IF_TEXT_EXTRACT_FAILS}":
ddgen_skip_row_if_extract_text_fails_fields =

    # Automatic renaming of tables
    # (Typical use: you make a view with a suffix "_x" as a working step, then
    # you want the suffix removed for users.)
ddgen_rename_tables_remove_suffixes =
    
    # Fields that are used as patient opt-out fields:
ddgen_patient_opt_out_fields =

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # DESTINATION INDEXING
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    # Fields to apply an index to
ddgen_index_fields =

    # Allow full-text index creation? Default true. Disable for databases that
    # don't support them?
ddgen_allow_fulltext_indexing = True

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # DATA DICTIONARY MANIPULATION TO DESTINATION TABLE/FIELD NAMES
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    # Force all destination tables/fields to lower case?
    # Boolean; default is True.
ddgen_force_lower_case = True

    # Convert spaces in table/fieldnames (yuk!) to underscores? Default: true.
ddgen_convert_odd_chars_to_underscore = True

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # PROCESSING OPTIONS, TO LIMIT DATA QUANTITY FOR TESTING
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    # Specify 0 (the default) for no limit, or a number of rows (e.g. 1000) to
    # apply to any tables listed in debug_limited_tables. For those tables,
    # only this many rows will be taken from the source database. Use this, for
    # example, to reduce the number of large documents fetched.
    # If you run a multiprocess/multithreaded anonymisation, this limit applies
    # per *process* (or task), not overall.
    # Note that these limits DO NOT APPLY to the fetching of patient-
    # identifiable information for anonymisation -- when a patient is
    # processed, all identifiable information for that patient is trawled.
debug_row_limit =

    # List of tables to which to apply debug_row_limit (see above).
debug_limited_tables =

# -----------------------------------------------------------------------------
# Source database example 2
# -----------------------------------------------------------------------------

[mysourcedb2]

url = mysql+mysqldb://username:password@127.0.0.1:3306/source2_databasename?charset=utf8

ddgen_force_lower_case = True
ddgen_per_table_pid_field = patient_id
ddgen_master_pid_fieldname = nhsnum
ddgen_table_blacklist =
ddgen_field_blacklist =
ddgen_table_require_field_absolute =
ddgen_table_require_field_conditional =
ddgen_pk_fields =
ddgen_constant_content = False
ddgen_scrubsrc_patient_fields =
ddgen_scrubsrc_thirdparty_fields =
ddgen_scrubmethod_code_fields =
ddgen_scrubmethod_date_fields =
ddgen_scrubmethod_number_fields =
ddgen_scrubmethod_phrase_fields =
ddgen_safe_fields_exempt_from_scrubbing =
ddgen_min_length_for_scrubbing = 4
ddgen_truncate_date_fields =
ddgen_filename_to_text_fields =
ddgen_binary_to_text_field_pairs =

# -----------------------------------------------------------------------------
# Source database example 3
# -----------------------------------------------------------------------------

[camcops]
# Example for the CamCOPS anonymisation staging database

url = mysql+mysqldb://username:password@127.0.0.1:3306/camcops_databasename?charset=utf8

# FOR EXAMPLE:
ddgen_force_lower_case = True
ddgen_per_table_pid_field = _patient_idnum1
ddgen_pid_defining_fieldnames = _patient_idnum1
ddgen_master_pid_fieldname = _patient_idnum2

ddgen_table_blacklist =

ddgen_field_blacklist = _patient_iddesc1
    _patient_idshortdesc1
    _patient_iddesc2
    _patient_idshortdesc2
    _patient_iddesc3
    _patient_idshortdesc3
    _patient_iddesc4
    _patient_idshortdesc4
    _patient_iddesc5
    _patient_idshortdesc5
    _patient_iddesc6
    _patient_idshortdesc6
    _patient_iddesc7
    _patient_idshortdesc7
    _patient_iddesc8
    _patient_idshortdesc8
    id
    patient_id
    _device
    _era
    _current
    _when_removed_exact
    _when_removed_batch_utc
    _removing_user
    _preserving_user
    _forcibly_preserved
    _predecessor_pk
    _successor_pk
    _manually_erased
    _manually_erased_at
    _manually_erasing_user
    _addition_pending
    _removal_pending
    _move_off_tablet

ddgen_table_require_field_absolute =
ddgen_table_require_field_conditional =
ddgen_pk_fields = _pk
ddgen_constant_content = False

ddgen_scrubsrc_patient_fields = _patient_forename
    _patient_surname
    _patient_dob
    _patient_idnum1
    _patient_idnum2
    _patient_idnum3
    _patient_idnum4
    _patient_idnum5
    _patient_idnum6
    _patient_idnum7
    _patient_idnum8

ddgen_scrubsrc_thirdparty_fields =

ddgen_scrubmethod_code_fields =
ddgen_scrubmethod_date_fields = _patient_dob
ddgen_scrubmethod_number_fields =
ddgen_scrubmethod_phrase_fields =

ddgen_safe_fields_exempt_from_scrubbing = _device
    _era
    _when_added_exact
    _adding_user
    _when_removed_exact
    _removing_user
    _preserving_user
    _manually_erased_at
    _manually_erasing_user
    when_last_modified
    when_created
    when_firstexit
    clinician_specialty
    clinician_name
    clinician_post
    clinician_professional_registration
    clinician_contact_details
# ... now some task-specific ones
    bdi_scale
    pause_start_time
    pause_end_time
    trial_start_time
    cue_start_time
    target_start_time
    detection_start_time
    iti_start_time
    iti_end_time
    trial_end_time
    response_time
    target_time
    choice_time
    discharge_date
    discharge_reason_code
    diagnosis_psych_1_icd10code
    diagnosis_psych_1_description
    diagnosis_psych_2_icd10code
    diagnosis_psych_2_description
    diagnosis_psych_3_icd10code
    diagnosis_psych_3_description
    diagnosis_psych_4_icd10code
    diagnosis_psych_4_description
    diagnosis_medical_1
    diagnosis_medical_2
    diagnosis_medical_3
    diagnosis_medical_4
    category_start_time
    category_response_time
    category_chosen
    gamble_fixed_option
    gamble_lottery_option_p
    gamble_lottery_option_q
    gamble_start_time
    gamble_response_time
    likelihood

ddgen_min_length_for_scrubbing = 4

ddgen_truncate_date_fields = _patient_dob
ddgen_filename_to_text_fields =
ddgen_binary_to_text_field_pairs =

""".format(  # noqa
    SCRUBSRC=SCRUBSRC,
    INDEX=INDEX,
    SCRUBMETHOD=SCRUBMETHOD,
    ALTERMETHOD=ALTERMETHOD,
    SRCFLAG=SRCFLAG,
    LONGTEXT=LONGTEXT,
    DEFAULT_MAX_ROWS_BEFORE_COMMIT=DEFAULT_MAX_ROWS_BEFORE_COMMIT,
    DEFAULT_MAX_BYTES_BEFORE_COMMIT=DEFAULT_MAX_BYTES_BEFORE_COMMIT,
    DECISION=DECISION,
    VERSION=VERSION,
    VERSION_DATE=VERSION_DATE,
)

# For the style:
#       [source_databases]
#       source1 = blah
#       source2 = thing
# ... you can't have multiple keys with the same name.
# http://stackoverflow.com/questions/287757



#!/usr/bin/env python
# crate_anon/nlp_manager/regex_func.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import typing.re
from typing import Any, Dict, Optional, Tuple

import regex
# noinspection PyProtectedMember
from regex import _regex_core

# =============================================================================
#  Core regex functions
# =============================================================================
# - All will use VERBOSE mode for legibility. (No impact on speed: compiled.)
# - Don't forget to use raw strings for all regex definitions!
# - Beware comments inside regexes. The comment parser isn't quite as benign
#   as you might think. Use very plain text only.
# - (?: XXX ) makes XXX into an unnamed group.


REGEX_COMPILE_FLAGS = (regex.IGNORECASE | regex.MULTILINE | regex.VERBOSE |
                       regex.UNICODE)


def at_wb_start_end(regex_str: str) -> str:
    """
    Caution using this. Digits do not end a word, so "mm3" will not match if
    your "mm" group ends in a word boundary.
    """
    return "\b(?: {} )\b".format(regex_str)


def at_start_wb(regex_str: str) -> str:
    """
    With word boundary at start. Beware, though; e.g. "3kg" is reasonable, and
    this does NOT have a word boundary in.
    """
    return "(?: \b (?: {} ) )".format(regex_str)


def compile_regex(regex_str: str) -> typing.re.Pattern:
    try:
        return regex.compile(regex_str, REGEX_COMPILE_FLAGS)
    except _regex_core.error:
        print("FAILING REGEX:\n{}".format(regex_str))
        raise


def compile_regex_dict(regexstr_to_value_dict: Dict[str, Any]) \
        -> Dict[typing.re.Pattern, Any]:
    return {
        compile_regex(k): v
        for k, v in regexstr_to_value_dict.items()
    }


def get_regex_dict_match(text: Optional[str],
                         regex_to_value_dict: Dict[typing.re.Pattern, Any],
                         default: Any = None) \
        -> Tuple[bool, Any]:
    """Returns (matched, result)."""
    if text:
        for r, value in regex_to_value_dict.items():
            if r.match(text):
                return True, value
    return False, default
